# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
''' This module managed ELB objects '''
from kloudi.utils import logging
from boto.ec2.elb import HealthCheck
from boto.exception import BotoServerError


class KloudiELB(object):
    ''' This class is used to manage ELBs '''
    # pylint:disable=too-many-arguments
    # pylint:disable=too-many-instance-attributes
    # Appropriate number of arguments for an ELB
    def __init__(self, conn, name, conf, verbose=False, find_only=False):
        self.elb = None
        self.created = False
        self.conn = conn
        self.name = name
        self.conf = conf
        self.azs_used = conf.get_azs()
        self.verbose = verbose
        self.find_only = find_only
        self.elb_listeners = [
            (conf.public_port, conf.private_port, conf.protocol)]
        self.init_elb()

    def init_elb(self):
        ''' Initialize ELB, create if does not exist by default '''
        try:
            self.elb = self.conn.get_all_load_balancers(
                load_balancer_names=[self.name])[0]
            logging.debug('Found existing ELB: %s' % self.name, self.verbose)
        except (IndexError, BotoServerError), msg:
            if self.find_only:
                logging.info('%s does not exist' % self.name)
                return False
        if self.elb is None:
            self.create_elb()
            self.set_hc()
            self.set_azs()

    def add_instances(self, instances):
        ''' Add instances to ELB '''
        logging.debug('Adding instances to ELB: %s' % instances, self.verbose)
        try:
            self.elb.register_instances(instances)
        except BotoServerError, msg:
            logging.error(msg.message)

    def remove_instances(self, instances):
        ''' Remove instances from ELB '''
        logging.debug(
            "Removing instances from ELB: %s" % instances, self.verbose)
        try:
            self.elb.deregister_instances(instances)
        except BotoServerError, msg:
            logging.error(msg.message)

    def set_azs(self):
        ''' Ensure AZs add to ELB from config '''
        azs = self.azs_used
        logging.debug("Trying to add AZs to ELB: %s" % azs, self.verbose)
        for zone in azs:
            if zone not in self.elb.availability_zones:
                logging.debug("Adding %s to ELB" % azs, self.verbose)
                try:
                    self.elb.enable_zones(zone)
                except BotoServerError, msg:
                    logging.error(msg.message)
        logging.info('Zones configured for ELB: %s' % self.azs_used)

    def set_hc(self):
        ''' Ensure HC is set for ELB '''
        hck = HealthCheck(
            self.conf.hc_access_point,
            self.conf.hc_interval,
            self.conf.hc_target,
            self.conf.hc_healthy_threshold,
            self.conf.hc_unhealthy_threshold)
        logging.info('Configuring HC: %s' % hck)
        try:
            self.elb.configure_health_check(hck)
        except BotoServerError, msg:
            logging.error(msg.message)

    def get_hc(self):
        ''' Return the healtcheck object '''
        try:
            return self.elb.health_check
        except BotoServerError, msg:
            logging.error(msg.message)
            return None

    def get_dns(self):
        ''' Return dns_name for ELB '''
        return self.elb.dns_name

    def check_elb_exists(self):
        ''' Check to see if ELB exists '''
        try:
            return self.conn.get_all_load_balancers(
                load_balancer_names=[self.name])
        except BotoServerError, msg:
            logging.error(msg.message)
            return None

    def get_instances(self):
        ''' Return instances assigned to ELB '''
        return self.elb.instances

    def get_azs(self):
        ''' Return AZs configured for ELB '''
        return self.elb.availability_zones

    def create_elb(self):
        ''' Create an ELB '''
        logging.debug('Create ELB %s' % self.name, self.verbose)
        try:
            self.elb = self.conn.create_load_balancer(
                self.name, self.azs_used, self.elb_listeners)
            logging.info('Created %s: %s' % (self.name, self.elb))
            self.created = True
        except BotoServerError, msg:
            logging.error(msg.message)

    def instance_configured(self, inst_id):
        ''' Check if instance is configured on ELB '''
        for instance in self.get_instances():
            if instance.id == inst_id:
                return True
        return False

    def destroy(self):
        ''' Destroy an ELB '''
        try:
            return self.elb.delete()
        except BotoServerError, msg:
            logging.error(msg.message)
            return None



# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
''' Logging interface for Kloudi '''
import sys
import datetime


def error(message):
    ''' Prints error messages to STDERR '''
    print >> sys.stderr, (
        '%s ERROR: %s' % (
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            message))


def warn(message):
    ''' Prints error messages to STDERR '''
    print >> sys.stderr, (
        '%s WARN: %s' % (
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            message))


def info(message):
    ''' Prints informational messages to STDOUT '''
    print >> sys.stdout, (
        '%s INFO: %s' % (
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            message))


def debug(message, verbose):
    ''' Prints verbose messaging to STDOUT '''
    if verbose:
        print >> sys.stdout, (
            '%s DEBUG: %s' % (
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                message))



#!/usr/bin/python2.7
''' Used to manage provision of EC2 servers and services. '''
# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import sys
import time
import optparse
import boto.sqs
import kloudi.utils.accounts
import kloudi.utils.naming
import kloudi.utils.tagger
import kloudi.utils.config
import kloudi.aws.ec2.elb
import kloudi.aws.ec2.ebs
import kloudi.aws.ec2.security_groups
from nvlib import Nventory
from zambi import ZambiConn
from kloudi.utils import logging
from boto.exception import SQSError
from boto.exception import EC2ResponseError
from kloudi.aws.ec2.instances import KloudiInstance


def parser_setup():
    ''' Setup the options parser. '''
    usage = 'usage: %prog [options]'
    desc = 'Provision instances and maintain SGs and ELBs'
    parser = optparse.OptionParser(usage=usage, description=desc)
    parser.add_option(
        '--config',
        '-c',
        action='store',
        type='string',
        dest='config_file',
        help='config (ini) file to use REQUIRED')
    parser.add_option(
        '--aws-account',
        '-A',
        action='store',
        type='string',
        dest='aws_account',
        help='AWS Account to use.')
    parser.add_option(
        '--instances',
        '-i',
        action='store',
        type='int',
        dest='count',
        help='number of instances to start')
    parser.add_option(
        '--verbose',
        '-v',
        action='store_true',
        default=False,
        dest='verbose',
        help='verbose output')
    parser.add_option(
        '--ami',
        '-a',
        action='store',
        type='string',
        dest='ami',
        help='AMI to use; should be a nv-configured name, '
        'but could be an actual AMI')
    parser.add_option(
        '--size',
        '-s',
        action='store',
        type='string',
        dest='size',
        help='instance size to use')
    parser.add_option(
        '--availability-zones',
        '-z',
        action='store',
        type='string',
        dest='azs',
        help='Availability Zone[s] (comma-separated) to use')
    parser.add_option(
        '--no-substitute-zones',
        '-Z',
        action='store_false',
        dest='substitute_zones',
        default=True,
        help='do not attempt another AZ on failure')
    parser.add_option(
        '--use-zones',
        '-u',
        action='store',
        type='int',
        dest='use_zones',
        default=None,
        help='maximum number of AZs to use')
    parser.add_option(
        '--allocate-eips',
        '-e',
        action='store_true',
        dest='allocate_eips',
        default=False,
        help='allocate and attach new EIPs for new instances')
    parser.add_option(
        '--nv-ini',
        '-n',
        action='store',
        type='string',
        dest='nv_ini',
        default="/app/secrets/nv_prd.ini",
        help='External DB ini file to use; useful for testing')
    parser.add_option(
        '--destroy',
        '-D',
        action='store',
        type='string',
        dest='destroy',
        default=None,
        help='set DESTROY to \'destroy\' to destroy the entire cluster, '
        'including ELB and SG')
    parser.add_option(
        '--hibernate',
        '-H',
        action='store_true',
        dest='hibernate',
        default=False,
        help='stop ALL instances in the cluster and exit')
    parser.add_option(
        '--wake',
        '-W',
        action='store_true',
        dest='wake',
        default=False,
        help='start ALL stopped instances in the cluster and exit')
    parser.add_option(
        '--status',
        '-S',
        action='store_true',
        dest='status',
        default=False,
        help='show the status of the cluster and exit')
    parser.add_option(
        '--retag',
        '-T',
        action='store_true',
        dest='retag',
        default=False,
        help='re-tag the cluster and exit')
    parser.add_option(
        '--skip-userdata',
        '-U',
        action='store_true',
        dest='skip_userdata',
        default=False,
        help='Skip Userdata script.')
    parser.add_option(
        '--ssh-key',
        '-k',
        action='store',
        type='string',
        dest='ssh_key',
        default=None,
        help='the ssh key ID used for starting instances')
    return parser


def check_timed_out(start_time, timeout, verbose=False):
    ''' Times out application if we are running to long. '''
    logging.debug(
        "check_timed_out(): Checking if %d - %d (%d) > %d" %
        (time.time(), start_time, time.time()-start_time, timeout), verbose)
    if time.time() - start_time > timeout:
        return True
    return False


def nv_connect(nv_ini):
    ''' Get a Nventory connection object. '''
    try:
        node_db = Nventory(ini_file=nv_ini)
        node_db.login()
    except Exception:
        raise
    return node_db


def cluster_tagger(conn, verbose=None):
    ''' Initialize the tagger. '''
    return kloudi.utils.tagger.Tagger(conn, verbose=verbose)


class KloudiEC2Provision(object):
    ''' This class manages provisioning in AWS. '''
    def __init__(self, opts):
        self.opts = opts
        self.cconfig = None
        self.conn = None
        self.hostnames = list()
        self.volume_instances = dict()
        self.volume_devices = dict()
        self.set_options()
        self.run_mode()

    def set_options(self):
        ''' Assign options and overrides for cluster configs. '''
        # Get the account name and pass to ClusterConfig
        if self.opts.aws_account:
            account_name = self.opts.aws_account
        else:
            account_name = kloudi.utils.config.get_account_from_conf(
                conf=self.opts.config_file)
        try:
            # Pass account name and load all the defaults
            self.cconfig = kloudi.utils.config.ClusterConfig(
                ini_file=self.opts.config_file,
                account_name=account_name,
                verbose=self.opts.verbose)
        except kloudi.utils.config.ConfigNotFoundException, msg:
            logging.error(msg)
            sys.exit(1)
        # this one MUST be "is not None", or a value of 0 won't work
        if self.opts.count is not None:
            logging.debug("setting count", self.opts.verbose)
            self.cconfig.set_count(self.opts.count)
        if self.opts.ami:
            logging.debug("setting AMI", self.opts.verbose)
            self.cconfig.set_ami(self.opts.ami)
        if self.opts.azs:
            logging.debug("setting AZs", self.opts.verbose)
            try:
                self.cconfig.set_azs(self.opts.azs)
            except kloudi.utils.config.InvalidAZError, msg:
                logging.error(msg)
                sys.exit(1)
        if self.opts.size:
            logging.debug("setting size", self.opts.verbose)
            self.cconfig.set_size(self.opts.size)
        if self.opts.use_zones:
            logging.debug("limiting AZs", self.opts.verbose)
            removed = self.cconfig.limit_azs(self.opts.use_zones)
            logging.debug("removed %d AZs" % removed, self.opts.verbose)
        if self.opts.ssh_key:
            logging.debug("setting ssh key", self.opts.verbose)
            self.cconfig.set_ssh_key(self.opts.ssh_key)
        if self.opts.allocate_eips:
            logging.debug("setting EIP allocation", self.opts.verbose)
            self.cconfig.set_allocate_eips()

    def run_mode(self):
        ''' Itterate through modes. '''
        if self.opts.destroy == "destroy":
            self.cluster_destroy()
        elif self.opts.destroy:
            logging.error(
                "Not a valid destory option: %s. Quitting." %
                self.opts.destroy)
            sys.exit(1)
        if self.opts.wake:
            self.cluster_wake()
        if self.opts.hibernate:
            self.cluster_hibernate()
        if self.opts.status:
            self.cluster_status()
        if self.opts.retag:
            self.cluster_retag()
        self.cluster_create()

    def aws_conn(self, service):
        ''' Gets a connection to EC2. '''
        zambi = ZambiConn()
        conn = zambi.get_connection(self.cconfig.get_aws_account(),
                                    service, self.cconfig.get_aws_region())
        return conn

    def cluster(self):
        ''' Gets an existing cluster object. '''
        node_db = nv_connect(self.opts.nv_ini)
        self.conn = self.aws_conn('ec2')
        try:
            cgc = kloudi.aws.ec2.instances.KloudiCluster(
                self.conn, name=self.cconfig.get_primary_sg(),
                node_db=node_db, verbose=self.opts.verbose)
        except kloudi.aws.ec2.instances.KloudiClusterNotFoundException, msg:
            logging.error("Problem finding cluster (%s)" % (msg))
            sys.exit(1)
        except kloudi.aws.ec2.instances.TooManySGsException, error:
            logging.error("Found multiple SGs! (%s)" % (error))
            sys.exit(1)
        except EC2ResponseError, msg:
            logging.error(msg.message)
            sys.exit(1)
        return cgc

    def elb_connection(self, find_only=True):
        ''' Get a connection to the ELB service. '''
        conn_elb = self.aws_conn('elb')
        try:
            kloudielb = kloudi.aws.ec2.elb.KloudiELB(
                conn_elb, self.cconfig.get_elb_name(),
                self.cconfig.get_elb_config(), find_only=find_only)
        except EC2ResponseError, msg:
            logging.error(msg.message)
        return kloudielb

    def cluster_destroy(self):
        ''' Destroy mode, delete all components for this cluster. '''
        logging.info("Tearing down %s in %s" % (
            self.cconfig.get_primary_sg(), self.cconfig.get_aws_region()))
        cgc = self.cluster()
        count = cgc.destroy()
        logging.info('Terminated %d instance(s)' % count)
        if self.cconfig.elb.enabled:
            kloudielb = self.elb_connection()
            if kloudielb.destroy():
                logging.info('ELB %s deleted' % kloudielb.name)
            else:
                logging.error('Deleting ELB %s failed')
        sgrp = kloudi.aws.ec2.security_groups.SecurityGroups(
            self.conn, self.cconfig.get_primary_sg(), find_only=True)
        if sgrp.destroy():
            logging.info("Security Group %s removed" % sgrp.name)
        logging.info('Tear down complete for %s' %
                     self.cconfig.get_primary_sg())
        sys.exit(0)

    def cluster_wake(self):
        ''' Wake a hibernating cluster. '''
        logging.info('Waking up %s' % self.cconfig.get_primary_sg())
        cgc = self.cluster()
        count = cgc.wake()
        logging.info("%d instance(s) in %s have been started" % (
            count, self.cconfig.get_primary_sg()))
        logging.info('Waking up instances complete')
        sys.exit(0)

    def cluster_hibernate(self):
        ''' Hibernates a running cluster. '''
        logging.info('Hibernating %s' % self.cconfig.get_primary_sg())
        cgc = self.cluster()
        count = cgc.hibernate()
        logging.info("%d instance(s) in %s have been hibernated" % (
            count, self.cconfig.get_primary_sg()))
        logging.info('Hibernating instances complete')
        sys.exit(0)

    def cluster_status(self):
        ''' Check the status of a cluster. '''
        logging.info('Checking status for %s' % self.cconfig.get_primary_sg())
        cgc = self.cluster()
        for instance in cgc.kloudiinstances:
            elbm = None
            elb_hc = None
            elb_azs = None
            ebs_vols = None
            try:
                kloudielb = self.elb_connection()
            except TypeError:
                kloudielb = None
            if kloudielb:
                if kloudielb.instance_configured(instance.inst_id):
                    elbm = kloudielb.get_dns()
                    elb_hc = kloudielb.get_hc()
                    elb_azs = kloudielb.get_azs()
            ebsm = instance.get_ebs_optimized()
            eipm = instance.get_associated_eip()
            vols = instance.get_non_root_volumes()
            if vols:
                ebs_vols = list()
                for key, value in vols.items():
                    ebs_vols.append('%s: %s' % (str(key), str(value)))
            msg = '''
            Instance %s
                ID: %s
                State: %s
                EBS Optimized: %s
                EBS Volumes: %s
                EIP: %s
                ELB %s
                    Health Check: %s
                    Availability Zones: %s
            ''' % (instance.name, instance.inst_id, instance.state,
                   ebsm, ebs_vols, eipm, elbm, elb_hc, elb_azs)
            logging.info(msg)
        logging.info('Status complete')
        sys.exit(0)

    def cluster_retag(self):
        ''' Retag the cluster. '''
        logging.info('Retagging cluster %s' % self.cconfig.get_primary_sg())
        cgc = self.cluster()
        tagger = cluster_tagger(self.conn, verbose=self.opts.verbose)
        if not tagger.add_tags(
                cgc.get_instance_ids(), self.cconfig.get_tagset()):
            logging.error('Problem addings tags')
            sys.exit(1)
        logging.info('Retag cluster complete')
        sys.exit(0)

    def check_config_types(self):
        ''' Check if there is an assigned ssh key. '''
        if self.cconfig.get_count() and not self.cconfig.get_ssh_key():
            logging.error(
                "You're trying to start instances, "
                "but don't have an SSH key set")
            sys.exit(1)
        node_db = nv_connect(self.opts.nv_ini)
        if not self.cconfig.get_resolved_ami(node_db):
            logging.error('Getting AMI failed, exiting')
            sys.exit(1)

    def sg_rules(self):
        ''' Find/create SG rules. '''
        primary_sg = kloudi.aws.ec2.security_groups.SecurityGroups(
            self.conn, self.cconfig.get_primary_sg())
        # add CIDR rules
        for rule in self.cconfig.get_cidr_rules():
            primary_sg.add_ingress(
                [rule['fport'], rule['lport']], rule['proto'],
                rule['cidr'])
        # add SG-to-SG rules
        for rule in self.cconfig.get_sg_rules():
            primary_sg.add_ingress(
                [rule['fport'], rule['lport']], rule['proto'],
                None, rule['owner'], rule['sg'])

    def userdata_replacements(self, host):
        ''' Replace properties to be passed to userdata. '''
        use_raid = self.cconfig.raid.enabled
        logging.debug("Determine if we need RAID", self.opts.verbose)
        if use_raid:
            raid_level = self.cconfig.raid.level
            logging.debug("Setting RAID level", self.opts.verbose)
            raid_device = self.cconfig.raid.device
            logging.debug("Setting RAID device", self.opts.verbose)
        else:
            raid_level = 0
            raid_device = "None"
            logging.debug('No RAID to configure', self.opts.verbose)
        fs_type = self.cconfig.get_fs_type()
        logging.debug('Setting fs_type: %s' % fs_type, self.opts.verbose)
        replacements = {
            '__HOSTNAME__': '"%s"' % host,
            '__DEVICES__': '%s' % len(self.cconfig.get_ebs_config()),
            '__USE_RAID__': '%s' % use_raid,
            '__RAID_LEVEL__': '%s' % raid_level,
            '__RAID_DEVICE__': '%s' % raid_device,
            '__FS_TYPE__': '%s' % fs_type,
            '__SKIP_USERDATA__': '%s' % self.opts.skip_userdata}
        logging.debug('Setting user Data Replacements: %s' %
                      replacements, self.opts.verbose)
        return replacements

    def cluster_create(self):
        ''' Provisions a new cluster based on a config. '''
        self.conn = self.aws_conn('ec2')
        node_db = nv_connect(self.opts.nv_ini)
        success = 0
        failed = 0
        self.check_config_types()
        logging.info('Applying SG Rules to %s' % self.cconfig.get_primary_sg())
        self.sg_rules()
        if self.cconfig.get_count():
            servers = dict()
            logging.debug(
                'Creating %d %s in %s using %s.' % (
                    self.cconfig.get_count(), self.cconfig.get_size(),
                    self.cconfig.get_azs(), self.cconfig.get_ami()),
                self.opts.verbose)
            self.hostnames = kloudi.utils.naming.find_available_hostnames(
                self.cconfig.get_primary_sg(), self.cconfig.get_count(),
                self.cconfig.get_aws_account(),
                self.cconfig.get_aws_region(), 'ctgrd.com', node_db)
            start_time = time.time()
            logging.debug(
                'Creating new servers: %s' % self.hostnames,
                self.opts.verbose)
            for host in self.hostnames:
                servers[host] = KloudiInstance(
                    conn=self.conn, node_db=node_db,
                    verbose=self.opts.verbose)
                userdata = self.cconfig.get_user_data(
                    self.userdata_replacements(host))
                tries = 1
                if self.opts.substitute_zones:
                    tries = len(self.cconfig.get_azs())
                while tries > 0:
                    tries -= 1
                    used_az = self.cconfig.get_next_az()
                    logging.info("Starting %s in %s" % (host, used_az))
                    instance = servers[host].start(
                        self.cconfig.get_ami(), self.cconfig.get_ssh_key(),
                        self.cconfig.get_sgs(), userdata,
                        host, self.cconfig.get_size(), used_az,
                        self.cconfig.get_node_groups(),
                        self.cconfig.get_allocate_eips(),
                        self.cconfig.get_use_ebs_optimized())
                    if instance:
                        success += 1
                        break
                    else:
                        if tries:
                            logging.warn(
                                'Failed to create %s in %s, retrying' %
                                (host, used_az))
                else:
                    logging.error(
                        "Failed to create %s in all AZs, trying next instance" %
                        host)
                    failed += 1
                if len(self.cconfig.get_ebs_config()) > 0:
                    self.create_ebs(used_az, host, servers[host].get_id())
            if failed == self.cconfig.get_count():
                logging.error(
                    '%d of %d failed to create, dying' %
                    (failed, self.cconfig.get_count()))
                sys.exit(1)
            logging.info(
                '%d of %d server(s) created' %
                (success, self.cconfig.get_count()))
            self.wait_for_servers(servers, start_time, success)
            if self.volume_instances:
                self.attach_ebs()
            self.tag_by_instance(servers)
            if self.cconfig.get_server_env() == 'prd':
                self.puppet_whitelist()
        logging.info('Cluster config complete')

    def create_ebs(self, used_az, host, instance_id):
        ''' Create new EBS volumes. '''
        cgebs = kloudi.aws.ec2.ebs.KloudiEBS(self.conn)
        for ebsv in self.cconfig.get_ebs_config():
            logging.info(
                "Creating EBS volume %s for %s" %
                (ebsv['device'], host))
            volume = cgebs.create_volume(
                ebsv['size'], used_az,
                ebsv['type'], ebsv['iops'])
            #pylint: disable=maybe-no-member
            self.volume_instances[volume.id] = instance_id
            self.volume_devices[volume.id] = ebsv['device']

    def wait_for_servers(self, servers, start_time, success):
        ''' Wait for servers that didn't fail to start to start up. '''
        started = 0
        failed = 0
        azs_used = list()
        logging.info('Waiting for %d server(s)to start' % success)
        stime = self.cconfig.get_sleep_step()
        for host in self.hostnames:
            status = servers[host].analyze_state()
            while status == 1:
                if check_timed_out(start_time,
                                   self.cconfig.get_launch_timeout(),
                                   verbose=self.opts.verbose):
                    logging.error(
                        '%s failed to start before time out' % host)
                    break
                logging.info('Wating for %s to enter running state' % host)
                time.sleep(stime)
                status = servers[host].analyze_state()
            if status == 0:
                started += 1
                if servers[host].get_az() not in azs_used:
                    azs_used.append(servers[host].get_az())
                logging.info('%s is running' % host)
            elif status == 2:
                failed += 1
        logging.debug(
            "%d started, %d failed, of %d total" %
            (started, failed, success), self.opts.verbose)
        if failed == success:
            logging.error(
                'All %d server(s) failed to start' % failed)
            sys.exit(1)
        elif started != success:
            logging.error(
                '%d started, %d failed to start' % (started, failed))
        else:
            logging.info(
                '%d of %d started' % (started, success))
        if self.cconfig.elb.enabled and azs_used:
            self.setup_elb(servers, azs_used)

    def setup_elb(self, servers, azs_used):
        ''' Createst the cluster ELB. '''
        self.cconfig.elb.set_azs(azs_used)
        kloudielb = self.elb_connection(find_only=False)
        for host in self.hostnames:
            try:
                instance_id = servers[host].get_id()
            except AttributeError, msg:
                logging.error(msg)
                instance_id = None
            if instance_id:
                kloudielb.add_instances([instance_id])

    def attach_ebs(self):
        ''' Attaches EBS volumes to instances. '''
        cgebs = kloudi.aws.ec2.ebs.KloudiEBS(self.conn)
        for vol_id in self.volume_instances:
            logging.info("Attaching EBS volume %s on %s to %s" % (
                vol_id, self.volume_instances[vol_id],
                self.volume_devices[vol_id]))
            cgebs.attach_volume(
                vol_id, self.volume_instances[vol_id],
                self.volume_devices[vol_id])
            cgebs.set_ebs_del_on_term(
                self.volume_instances[vol_id], self.volume_devices[vol_id])

    def tag_by_instance(self, servers):
        ''' Tag resources tied to an instnace ID. '''
        tagger = cluster_tagger(self.conn, verbose=self.opts.verbose)
        for host in self.hostnames:
            try:
                instance_id = servers[host].get_id()
                if not tagger.add_tags(
                        [instance_id], self.cconfig.get_tagset()):
                    logging.error("Problem adding %s to %s" % (
                        self.cconfig.get_tagset(), instance_id))
            except AttributeError:
                instance_id = None
                logging.warn(
                    'Failed to set cost tags on failed '
                    'instance %s' % host)

    def puppet_whitelist(self):
        ''' Add the puppet whitelists. '''
        conn_sqs = self.aws_conn('sqs')
        qurl = self.cconfig.get_whitelist_url()
        try:
            sqsq = boto.sqs.queue.Queue(conn_sqs, url=qurl)
        except SQSError, msg:
            logging.error(msg.message)
            sys.exit(1)
        for host in self.hostnames:
            try:
                #pylint: disable=maybe-no-member
                conn_sqs.send_message(sqsq, host)
            except SQSError, msg:
                logging.error(msg.message)


def main():
    ''' Read in command line options and manage instance provisioning. '''
    parser = parser_setup()
    (opts, args) = parser.parse_args()
    if args:
        parser.print_help()
        parser.error('Too many arguments')
    if opts.config_file is None:
        parser.print_help()
        parser.error('"-c CONFIG_FILE" is required')
    KloudiEC2Provision(opts)


if __name__ == '__main__':
    main()



''' This module is used for interfacing with AWS S3 '''
# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import kloudi.utils.tagger


class KloudiS3Bucket(object):
    ''' This class is used to interface with S3 '''
    def __init__(self, conn, bucket):
        self.conn = conn
        self.bucket = bucket
        if self.bucket_lookup() is False:
            self.create_bucket()

    def bucket_lookup(self):
        ''' Lookup to see if bucket exists '''
        try:
            if self.conn.lookup(self.bucket):
                return True
            else:
                return False
        except Exception:
            raise

    def create_bucket(self):
        ''' Create an S3 Bucket '''
        try:
            self.conn.create_bucket(self.bucket)
        except Exception:
            raise

    def delete_bucket(self):
        ''' Delete an S3 bucket '''
        try:
            self.conn.delete_bucket(self.bucket)
        except Exception:
            raise

    def get_bucket(self):
        ''' Returns an S3 bucket object '''
        try:
            s3_bucket = self.conn.get_bucket(self.bucket)
            return s3_bucket
        except Exception:
            raise

    def upload_policy(self, data):
        ''' Set and replace a S3 bucket policy '''
        s3_bucket = self.get_bucket()
        try:
            s3_bucket.set_policy(data)
        except Exception:
            raise

    def set_tags(self, tagset, verbose):
        ''' Set the cost tags for a bucket '''
        try:
            tagger = kloudi.utils.tagger.Tagger(self.conn, verbose=verbose)
            tagger.add_tags([self.bucket], tagset)
        except Exception:
            raise



# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
''' Manages AWS security groups '''
from time import sleep
from boto.exception import EC2ResponseError
from kloudi.utils import logging


class SecurityGroups(object):
    ''' Class to manage Security Groups '''
    def __init__(self, conn, name, find_only=False, verbose=False):
        self.conn = conn
        self.name = name
        self.sgrp = None
        self.find_only = find_only
        self.verbose = verbose
        try:
            self.sgrp = conn.get_all_security_groups(name)[0]
        except (IndexError, EC2ResponseError), msg:
            if not self.find_only:
                logging.info('Creating SG %s' % self.name)
                self.sgrp = self.create()
            else:
                logging.error(msg.message)

    def create(self):
        ''' Creates a new Security Group '''
        desc = '%s Kloudi SG auto group' % self.name
        try:
            return self.conn.create_security_group(self.name, desc)
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None

    def add_ingress(self, port_pair, protocol, src_cidr=None,
                    src_acct=None, src_sg=None):
        ''' Add Ingress rule '''
        # pylint: disable=too-many-arguments
        # Appropriate number arguments of for add_ingress
        if src_cidr:
            try:
                self.sgrp.authorize(protocol, port_pair[0],
                                    port_pair[1], src_cidr)
                return True
            except EC2ResponseError, msg:
                if msg.error_code == 'InvalidPermission.Duplicate':
                    logging.debug(msg.error_message, self.verbose)
                else:
                    logging.error(msg.error_message)
                    return False
        elif src_acct and src_sg:
            try:
                self.conn.authorize_security_group(
                    self.name, src_sg, src_acct,
                    protocol, port_pair[0], port_pair[1], None)
                return True
            except EC2ResponseError, msg:
                if msg.error_code == 'InvalidPermission.Duplicate':
                    logging.debug(msg.error_message, self.verbose)
                else:
                    logging.error(msg.error_message)
                    return False
        else:
            logging.error(
                'add_ingress(port_pair=%s, protocol=%s, src_cidr=%s, '
                'rc_acct=%s, src_sg=%s) FAILED (missing data?)' %
                (port_pair, protocol, src_cidr, src_acct, src_sg))
            return False

    def destroy(self):
        ''' Destroys a Security Group '''
        stime = 10
        timeout = 120
        while timeout > 0:
            try:
                self.sgrp.delete()
                return True
            except EC2ResponseError:
                logging.warn(
                    "SG %s could not be removed, sleeping %ds" %
                    (self.name, stime))
                sleep(stime)
                timeout -= stime
        logging.error('SG %s could not be deleted')



# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
''' Generates JSON Statements for AWS policies '''
import sys
import ipaddr
import kloudi.utils.naming
from kloudi.utils import logging


def make_statement(user_acct, user, path, action,
                   effect, condition):
    ''' Generate JSON statement '''
    statement = {}
    statement['Action'] = do_action(action)
    statement['Resource'] = do_resource(path)
    statement['Effect'] = effect
    statement['Principal'] = do_principal(user_acct, user)
    if condition != 'empty':
        statement['Condition'] = do_condition(condition)
    return statement


def do_action(action):
    ''' Returns actions in JSON '''
    return action.split(',')


def do_principal(user_account, user):
    ''' Returns prinipcal in JSON '''
    item = dict()
    if user_account == 'cidr-networks':
        item['AWS'] = '*'
    elif user == 'root':
        item['AWS'] = (['arn:aws:iam::' +
                        str(user_account) +
                        ':root'])
    else:
        item['AWS'] = (['arn:aws:iam::' +
                        str(user_account) +
                        ':user/' + user])
    return item


def do_condition(condition):
    ''' Generate the condition block in the json
    condition is a three item string like:
    ConditionName,ConditionProperty,ConditionValue '''
    item = dict()
    if condition and len(condition.split(',')) == 3:
        (name, prop, value) = condition.split(',')
        item[name] = {}
        if is_ipnetwork(value):
            item[name][prop] = value
        elif kloudi.utils.naming.get_cidr(value):
            value = kloudi.utils.naming.get_cidr(value)
            item[name][prop] = value
        else:
            item[name][prop] = value
        return item
    elif condition == 'empty':
        pass
    else:
        logging.warn('Not enough values given to assign '
                              'condition in %s' % condition)
        return False


def do_resource(path):
    ''' Generate the resource block in the json (list of paths) '''
    resource = []
    resource.append('arn:aws:s3:::%s' % path)
    return resource


def is_ipnetwork(value):
    ''' Checks wether a given value is a valid IPv4 network, returns boolean '''
    try:
        ipaddr.IPNetwork(value)
        return True
    except ValueError:
        return False



#!/usr/bin/python2.7
# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
''' A tool that generates JSON policy objects from config files '''
import json
import os
import sys
import optparse
import kloudi.utils.config
from kloudi.utils.jgp import gen_entry
from kloudi.utils.jgp import statement
from kloudi.utils import accounts
from kloudi.aws.s3.bucket import KloudiS3Bucket
from zambi import ZambiConn


def gen_statement(entries):
    ''' Generates S3 bucket policy statments '''
    statements = []
    data = {}
    for entry in entries:
        (effect, action, user,
         user_acct, path, condition) = split_parameter(entry)
        statements.append(statement.make_statement(user_acct, user,
                                                   path, action,
                                                   effect, condition))
    data['Statement'] = statements
    return data


def to_json(data):
    ''' Converts data to JSON '''
    try:
        return json.dumps(data, indent=4, separators=(',', ': '))
    except Exception:
        raise


def split_parameter(entry):
    ''' Splits entry into multiple fields '''
    field = entry.split('|')
    return (field[0], field[1], field[2], field[3], field[4], field[5])


def generate_file(data, output_dir, account, bucket):
    ''' Dump JSON policy to a file '''
    json_file = ('%s/%s-%s.json' % (output_dir, account, bucket))
    print 'INFO: Saving JSON ouput to %s' % json_file
    conf = open(json_file, 'w')
    conf.write(data)
    conf.close()


def send_to_aws(data, account, bucket, tagset, verbose):
    ''' Creates a connection to AWS and sends data '''
    conn_manager = ZambiConn()
    conn = conn_manager.get_connection(account, service='s3')
    s3_bucket = KloudiS3Bucket(conn, bucket)
    print 'INFO: Updating %s:%s' % (account, bucket)
    print 'INFO: Setting tags'
    if verbose:
        print 'INFO: TAGS:\n%s' % tagset
    s3_bucket.set_tags(tagset, verbose)
    if data:
        print 'INFO: Setting bucket policy'
        if verbose:
            print 'INFO: JSON OUTPUT:\n%s' % data
            s3_bucket.upload_policy(data)
        else:
            s3_bucket.upload_policy(data)


def generate_entries(user, config):
    ''' Generates entries from config files '''
    ini = gen_entry.read_config(config)
    if 'cidr-networks' in user:
        acct_id = 'cidr-networks'
        user = 'cidr-networks'
    else:
        (user_acct, user) = user.split('-', 1)
        if user_acct.isdigit():
            acct_id = user_acct
        else:
            acct_id = accounts.get_account_id(user_acct)
    return gen_entry.gen_s3_entry(ini, user, acct_id)


def get_tags(config, verbose):
    ''' Get bucket cost tags from config '''
    cluster_config = kloudi.utils.config.ClusterConfig(
        ini_file=config,
        prv_type='s3',
        verbose=verbose,
        no_defaults=True)
    return cluster_config.get_tagset()


def parser_setup():
    ''' Setup the options parser '''
    usage = 'usage: %prog [options]'
    desc = 'Generate S3 bucket json policy'
    parser = optparse.OptionParser(usage=usage, description=desc)
    parser.add_option('--config-dir',
                      '-c',
                      action='store',
                      dest='config_dir',
                      type='string',
                      help='Policy config directory location')
    parser.add_option('--bucket',
                      '-b',
                      action='store',
                      dest='acct_bucket',
                      type='string',
                      help='S3 Bucket to update')
    parser.add_option('--output-dir',
                      '-o',
                      action='store',
                      dest='output_dir',
                      type='string',
                      help='Output dir to save json file for bucket')
    parser.add_option('--verbose',
                      '-v',
                      action='store_true',
                      default=False,
                      dest='verbose',
                      help='Enables verbose output')
    return parser


def main():
    ''' The main program executed '''
    parser = parser_setup()
    (options, args) = parser.parse_args()
    if args:
        parser.print_help()
        parser.error('Too many arguments')
    if options.config_dir is None or options.acct_bucket is None:
        parser.print_help()
        parser.error('CONFIG-DIR and BUCKET must be specified')
    (account, bucket) = options.acct_bucket.split('/')
    conf_files = next(os.walk(options.config_dir +
                              '/' + options.acct_bucket))[2]
    entries = list()
    for conf in conf_files:
        config = (options.config_dir + '/' + options.acct_bucket + '/' + conf)
        user = conf.split('.ini')[0]
        if user:
            if user == 'meta':
                tagset = get_tags(config, options.verbose)
            else:
                entries += generate_entries(user, config)
    if len(entries) == 0:
        policy = None
    else:
        policy = to_json(gen_statement(entries))
    if 'tagset' not in vars():
        print 'ERROR: Missing tags in meta.ini'
        sys.exit(1)
    if options.output_dir is None:
        send_to_aws(policy, account, bucket, tagset, options.verbose)
    else:
        generate_file(policy, options.output_dir, account, bucket)


if __name__ == '__main__':
    main()



# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
''' Returns entries read in from config files to be used
to generates statement policies'''
import re
import sys
import ConfigParser
from kloudi.utils import logging


def read_config(config):
    ''' Read in config ini file for policies '''
    ini = ConfigParser.ConfigParser()
    ini.read(config)
    if ini.sections():
        return ini
    else:
        return False


def gen_s3_entry(ini, user, user_account):
    ''' Generate entries from config '''
    entry = list()
    pattern = re.compile("path")
    if ini.sections():
        for action in ini.sections():
            try:
                effect = ini.get(action, 'effect')
            except ConfigParser.NoOptionError, msg:
                logging.warn(msg)
            if ini.has_option(action, 'condition'):
                condition = ini.get(action, 'condition')
            else:
                condition = 'empty'
            for item in ini.items(action):
                if pattern.match(item[0]):
                    path = item[1]
                    entry.append('%s|%s|%s|%s|%s|%s' % (effect, action, user,
                                                        user_account, path,
                                                        condition))
    return sorted(entry)



# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
'''
Library for CGM naming conventions
'''

import os
import re
import sys
import kloudi.utils.accounts
from nvlib import Nventory
from kloudi.utils import logging


def find_available_hostnames(group, count=1, account=None,
                             region=None, domain=None, node_db=None):
    '''
    Return the first <count> appropriate hostnames not already in node_db
    '''
    retnodes = list()
    nodes = list()
    if node_db is None:
        node_db = Nventory(url='http://fakehost')
    if account is None:
        return False
    data = {
        'loc': get_aws_dc(region),
        'env': group[:3],
        'sclass': group[3:],
        'acct': account,
        'domain': domain}
    node_data = node_db.get_nodes(data)
    if node_data:
        for node in node_data:
            nodes.append(node['name'])
    num = 1
    while len(retnodes) != count:
        possible = gen_hostname(group, num, account, region, domain)
        if possible not in nodes:
            retnodes.append(possible)
        num += 1
    return retnodes


def gen_hostname(group, num, account=None, region="us-east-1", domain=None):
    ''' Return the CGM hostname if the SG uses the naming convention '''
    if domain is None:
        domain = "ctgrd.com"
    if re.search("^(prd|dev|prg|qat|sts|tst|utl|stg|sbx)[a-z][a-z][a-z]$",
                 group):
        return ("%s%s%d.%s.%s" %
                (get_aws_dc(region), group, num, account, domain))
    else:
        return False


def get_aws_dc(region):
    ''' Returns the CGM standard AWS datacenter. '''
    if re.match('us-east-1', region):
        return 'aws1'
    elif re.match('us-west-1', region):
        return 'aws2'
    elif re.match('eu-west-1', region):
        return "aws3"
    else:
        return False


def get_logging_bucket_name(account_id=None):
    '''
    Get the standardized bucket name for account_id or the current env account
    '''
    account_name = kloudi.utils.accounts.get_account_name(
        account_id=account_id)
    if account_name:
        return "cgs3log-%s" % account_name
    else:
        return False


def get_cidr(net):
    ''' Get the CIDR address from a network name '''
    names = get_network_data()[1]
    try:
        return names[net]
    except KeyError:
        pass


def get_network_data():
    ''' Parse the network file for use by other methods '''
    nets = dict()
    names = dict()
    net_file = '%s/%s' % (os.getenv('AWS_CONF_DIR'), '/networks.txt')
    try:
        nfile = open(net_file, 'r')
    except IOError, msg:
        logging.error(msg)
        return False
    for line in nfile.readlines():
        ent = line.strip().split(":")
        nets[ent[0]] = ent[1]
        names[ent[1]] = ent[0]
    return nets, names



# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
''' This module Manages tagging '''
import os
import boto.s3.tagging
from kloudi.utils import logging
from boto.exception import EC2ResponseError
from boto.exception import S3ResponseError


def get_validation_file(tname):
    ''' Return cost tags validation file '''
    return os.getenv('AWS_CONF_DIR') + '/cost_tags.' + tname

def validate_tag(tname, tvalue):
    ''' Check if the tag value is already allowed to avoid typos '''
    allowed = list()
    tag_file = get_validation_file(tname)
    # Component is an optional tag
    if tname == 'Component':
        return True
    try:
        tnfile = open(tag_file, 'r')
    except IOError, msg:
        logging.error(msg)
    for line in tnfile.readlines():
        allowed.append(line.strip())
    return tvalue in allowed


class Tagger(object):
    ''' Kloudi tagger object '''
    def __init__(self, conn, tagnames=None, verbose=False):
        self.conn = conn
        self.verbose = verbose
        self.prefix_cost = '/costs_tags.'
        if tagnames:
            self.tagnames = tagnames
        else:
            self.tagnames = [
                'BusinessUnit', 'Team', 'Project', 'Env', 'Component']

    def get_cost_tags_health(self, rid, tag_type=None):
        ''' Return cost tags for given resource '''
        tags = dict()
        if tag_type == "ec2" or rid[:2] == 'i-':
            instances = self.conn.get_all_instances([rid])
            instance = instances[0].instances[0]
            tags = instance.tags
        elif tag_type == 'ebs' or rid[:4] == 'vol-':
            vol = self.conn.get_all_volumes([rid])[0]
            tags = vol.tags
        return self.get_cost_tags_health_from_tags(tags)

    def get_cost_tags_health_from_tags(self, tags):
        ''' Check to ensure all required cost tags are healthy '''
        ret = dict()
        healthy = True
        for key in self.tagnames:
            try:
                ret[key] = tags[key]
            except KeyError, msg:
                logging.error('No key %s in tags' % msg)
                healthy = False
                ret[key] = None
        return healthy, ret

    def get_cost_tags(self, rid, tag_type=None):
        ''' Return cost tags '''
        healthy, ret = self.get_cost_tags_health(rid, tag_type=tag_type)
        logging.info('Retrived tags: %s Healhty: %s' % (ret, healthy))
        return ret

    def add_tags(self, rids, tagset, tag_type=None):
        ''' Set one or more tags on a list of IDs '''
        failed = 0
        for rid in rids:
            logging.debug('Tagging resource: %s' % rid, self.verbose)
            if not self._add_tags(rid, tagset, tag_type=tag_type):
                failed += 1
        if not failed:
            return True
        else:
            return False

    def _add_tags(self, rid, tagset, tag_type=None):
        ''' Set one or more tags on a single ID '''
        failed = 0
        if tag_type == 'ec2' or rid[:2] == 'i-':
            rid = self.conn.get_all_instances([rid])
            instance = rid[0].instances[0]
            for tname, tvalue in tagset.items():
                logging.info('For %s adding, %s: %s' %
                             (instance, tname, tvalue))
                try:
                    instance.add_tag(tname, tvalue)
                except EC2ResponseError, msg:
                    logging.error(msg.message)
                    failed += 1
            logging.info('Checking for attached EBS volumes')
            try:
                volumes = self.conn.get_all_volumes()
            except EC2ResponseError, msg:
                logging.error(msg.message)
                failed += 1
            for vol in volumes:
                if vol.attach_data.instance_id == instance.id:
                    logging.info('Found attached vol: %s' % vol.id)
                    self._add_tags(vol.id, tagset, tag_type='ebs')
        elif tag_type == 'ebs' or rid[:4] == 'vol-':
            try:
                vol = self.conn.get_all_volumes([rid])[0]
            except EC2ResponseError, msg:
                logging.error(msg.message)
                failed += 1
            for tname, tvalue in tagset.items():
                logging.info('For %s adding, %s:%s' % (vol.id, tname, tvalue))
                try:
                    vol.add_tag(tname, tvalue)
                except EC2ResponseError, msg:
                    logging.error(msg.message)
                    failed += 1
        elif tag_type == 'rds':
            return True
        else:
            return self.tag_s3_bucket(rid, tagset)
        if not failed:
            return True
        else:
            return False

    def tag_s3_bucket(self, rid, tagset):
        ''' Tags S3 buckets with complicated s3 tagging shenanigans '''
        new_tags = list()
        failed = 0
        logging.info('Tagging S3 bucket %s' % rid)
        try:
            bucket = self.conn.get_bucket(rid)
        except S3ResponseError, msg:
            logging.error(msg.message)
            failed += 1
        try:
            logging.debug('Check for existing tags...', self.verbose)
            existing_tags = bucket.get_tags()[0]
        except S3ResponseError, msg:
            logging.error(msg.message)
            existing_tags = dict()
        logging.debug('Existing tags: %s' % existing_tags.to_xml(), self.verbose)
        try:
            tags = boto.s3.tagging.Tags()
            tset = boto.s3.tagging.TagSet()
        except S3ResponseError, msg:
            logging.error(msg.message)
            failed += 1
        for tname, tvalue in tagset.items():
            logging.debug('%s:%s is a new tag' % (tname, tvalue), self.verbose)
            try:
                tset.add_tag(tname, tvalue)
                new_tags.append(tname)
            except S3ResponseError, msg:
                logging.error(msg.message)
                failed += 1
        for tag in existing_tags:
            if tag.key not in new_tags:
                logging.info('Tagging %s, %s: %s' %
                             (rid, tag.key, tag.value))
                try:
                    tset.add_tag(tag.key, tag.value)
                    new_tags.append(tag.key)
                except S3ResponseError, msg:
                    logging.error(msg.message)
                    failed += 1
        logging.info('Submitting tagset for S3 bucket')
        try:
            tags.add_tag_set(tset)
            bucket.set_tags(tags)
        except S3ResponseError, msg:
            logging.error(msg.message)
            failed += 1
        if not failed:
            return True
        else:
            return False



# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
''' This module manages EC2 clusters and instance objects '''
import time
import socket
from kloudi.utils import logging
from boto.exception import EC2ResponseError


def wait_for_instance(instance, desired_state="up", timeout=120, verbose=False):
    ''' Waits for instance to enter desired state. '''
    logging.debug('Instance: %s, Desired State: %s' %
                  (instance.name, desired_state), verbose)
    stt = time.time()
    while time.time() - stt < timeout:
        time.sleep(10)
        state = instance.analyze_state(desired_state)
        if state == 0:
            logging.debug('Instance entered desired state: %s' %
                          desired_state, verbose)
            return True
        elif state == 2:
            logging.error('Instance failed to enter desired state: %s' %
                          desired_state)
            return False
    logging.error(
        'Waiting for %s timed out after %ds' % (instance.name, timeout))
    return False


class KloudiCluster(object):
    ''' The cluster specifc to ct_env and ct_class. '''
    def __init__(self, conn, name=None, node_db=None, verbose=None):
        self.conn = conn
        self.name = name
        self.node_db = node_db
        self.verbose = verbose
        self.instances = list()
        self.kloudiinstances = list()
        if self.name is not None:
            self.get_instances()

    def get_instances(self, instance_ids=None):
        ''' Get both instances and Kloudiinstances '''
        if instance_ids:
            all_instances = None
            instances = list()
            try:
                all_instances = self.conn.get_all_instances(
                    instance_ids=instance_ids)
            except EC2ResponseError, msg:
                logging.error(msg.message)
                return None
            for instance in all_instances:
                instances.append(instance.instances[0])
            self.instances = instances
        else:
            try:
                sgrp = self.conn.get_all_security_groups(self.name)[0]
            except IndexError:
                logging.error('No instances found: check name %s' % self.name)
                return None
            try:
                self.instances = sgrp.instances()
            except EC2ResponseError, msg:
                logging.error(msg.message)
        # we should really use KloudiInstances for everything
        for inst in self.instances:
            self.kloudiinstances.append(KloudiInstance(
                self.conn, inst_id=inst.id,
                node_db=self.node_db, verbose=self.verbose))
        return self.instances

    def get_instance_ids(self):
        ''' Returns instance IDs '''
        ids = list()
        for iid in self.instances:
            ids.append(iid.id)
        return ids

    def add_instance(self, kloudiinstance):
        ''' Add instance to cluster. '''
        self.kloudiinstances.append(kloudiinstance)

    def destroy(self):
        ''' Terminates instances in cluster. '''
        count = 0
        for iid in self.kloudiinstances:
            if iid.get_state() not in ['terminated']:
                if iid.destroy():
                    logging.info(
                        'Waiting for %s (%s) to terminate' %
                        (iid.name, iid.inst_id))
                    if wait_for_instance(iid, desired_state='down',
                                         verbose=self.verbose):
                        count += 1
            else:
                logging.warn('%s already teriminated' % iid.name)
                count += 1
        if count != len(self.kloudiinstances):
            logging.warn(
                'Asked for %d but only %d terminated' %
                (len(self.kloudiinstances), count))
        return count

    def hibernate(self):
        ''' Hibernate instances in cluster. '''
        count = 0
        for iid in self.kloudiinstances:
            if iid.hibernate():
                logging.info('Waiting for %s to stop' % iid.name)
                if wait_for_instance(iid, desired_state='down',
                                     verbose=self.verbose):
                    count += 1
        if count != len(self.kloudiinstances):
            logging.warn(
                'Asked for %d but only %d stopped' %
                (len(self.kloudiinstances), count))
        return count

    def wake(self):
        ''' Wake instances in cluster. '''
        count = 0
        for iid in self.kloudiinstances:
            if iid.wake():
                logging.info('Waiting for %s to start' % iid.name)
                if wait_for_instance(iid, verbose=self.verbose):
                    logging.debug('Wait for %s successful' %
                                  iid.name, self.verbose)
                    count += 1
        if count != len(self.kloudiinstances):
            logging.warn(
                'Asked for %d but only %d started' %
                (len(self.kloudiinstances), count))
        return count


class KloudiInstance(object):
    ''' Class that  manages instance objects '''
    # pylint:disable=too-many-instance-attributes
    # Required for boto API
    def __init__(self, conn, inst_id=None, node_db=None, verbose=False):
        self.conn = conn
        self.inst_id = inst_id
        self.verbose = verbose
        self._instance = None
        self._reservation = None
        self.start_time = None
        self.registered = False
        self.node_db = node_db
        self.allocateeips = None
        self.state = None
        self.name = None
        self.reeip = None
        self.eip = None
        if inst_id:
            self._instance = conn.get_all_instances([inst_id])[0].instances[0]
            self.name = self._instance.tags.get("Name") or self.inst_id
            self.state = self._instance.state

    def start(self, ami, sshkey, sgs, user_data, hostname,
              isize, zone, nodegroups, allocateeips, use_ebsoptimized):
        ''' Starts an EC2 instance '''
        # pylint:disable=too-many-arguments
        # Required for boto API
        logging.debug(
            'KloudiInstance.start(%s, %s, %s, %s, %s, %s, %s, %s)' %
            (ami, sshkey, sgs, len(user_data), hostname, isize, zone,
             nodegroups), self.verbose)
        try:
            self._reservation = self.conn.run_instances(
                ami, 1, 1, sshkey, sgs, user_data, None, isize, zone,
                None, None, False, None, None, ebs_optimized=use_ebsoptimized)
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None
        self._instance = self._reservation.instances[0]
        self.inst_id = self._instance.id
        self.name = hostname
        self.allocateeips = allocateeips
        safety = 10
        while safety:
            try:
                self._instance.add_tag('Name', hostname)
                break
            except EC2ResponseError, msg:
                logging.error(msg.message)
            safety -= 1
            time.sleep(1)
        if self.node_db:
            self.node_register(self._instance.id, hostname)
            if nodegroups:
                self.add_node_groups(self._instance.id, nodegroups)
        self.start_time = time.time()
        return self._instance.id

    def node_register(self, inst_id, hostname):
        ''' Try to register the new instance with node_db '''
        return self.node_db.register_host(hostname, inst_id)

    def add_node_groups(self, inst_id, nodegroups):
        ''' Try to register the new instance with node_db '''
        return self.node_db.add_node_groups(inst_id, nodegroups)

    def set_state(self, status):
        ''' Set nVentory state for instance '''
        return self.node_db.set_status(self.inst_id, status)

    def get_id(self):
        ''' Return the EC2 Instance ID '''
        return self._instance.id

    def get_non_root_volumes(self):
        ''' Returns a list of non root ebs volumes attached to the instance. '''
        vols = dict()
        inst = self._instance
        if inst.block_device_mapping:
            for block in inst.block_device_mapping:
                if '/dev/sda1' not in block:
                    vols[block] = inst.block_device_mapping[block].volume_id
        return vols

    def get_ebs_optimized(self):
        ''' Return ebs optimized option '''
        return self._instance.ebs_optimized

    def get_az(self):
        ''' Return the AZ '''
        return self._instance.placement

    def get_state(self):
        ''' Update the state and return '''
        self.state = self._instance.update()
        return self.state

    def analyze_state(self, desired_state='up'):
        ''' Find out if we're done (0), waiting (1), or screwed (2) '''
        result = None
        state = self.get_state()
        if state:
            logging.debug(
                'Analyze state %s is %s' % (self._instance.id, state),
                self.verbose)
            pending_up = ['pending']
            pending_down = ['shutting-down', 'stopping']
            ec2_up = 'running'
            ec2_down = ['terminated', 'stopped']
            if desired_state is 'up':
                if state in pending_up:
                    result = 1
                elif state in pending_down:
                    result = 2
                elif state == ec2_up:
                    self.finalize_start()
                    result = 0
            else:
                if state in pending_down:
                    result = 1
                elif state == ec2_up:
                    result = 2
                elif state in ec2_down:
                    result = 0
            return result
        else:
            logging.error('Unable to get status for %s' % self._instance.id)
            return False

    def finalize_start(self):
        ''' Perform EC2 actions once the instance is finally 'running' '''
        logging.debug(
            'Finalize start: %s' % self.inst_id, self.verbose)
        if self.allocateeips:
            return self.new_eip()
        if self.reeip:
            logging.debug('Reassociating EIP %s with instance %s' %
                          (self.reeip, self.inst_id), self.verbose)
            try:
                self.reeip.associate(self.inst_id)
                self.reeip = None
                return True
            except EC2ResponseError, msg:
                logging.error(msg.message)
                return None
        return True

    def get_eip_by_addr(self, myip):
        ''' Return an EIP address '''
        try:
            eips = self.conn.get_all_addresses()
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None
        for eip in eips:
            if myip == eip.public_ip:
                return eip
        return None

    def set_eip(self, address):
        ''' Set the EIP for the instance '''
        try:
            self.eip = self.get_eip_by_addr(address)
            return True
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None

    def get_eip(self, steal=False):
        ''' Figure out if my hostname is an EIP '''
        if self.eip:
            return self.eip
        data = self.node_db.get_node_by_instance_id(self.inst_id)[0]
        try:
            myip = socket.gethostbyname(data['ec2_public_hostname'])
        except socket.gaierror, msg:
            logging.error(msg)
            return None
        logging.debug('myip: %s' % myip, self.verbose)
        eip = self.get_eip_by_addr(myip)
        if eip:
            logging.debug('EIP: %s associated with %s' % (eip, eip.instance_id),
                          self.verbose)
            if steal or not eip.instance_id:
                return eip
        return None

    def get_associated_eip(self):
        ''' Return EIP associated with EC2 instance '''
        try:
            eips = self.conn.get_all_addresses()
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None
        if eips:
            for eip in eips:
                if eip.instance_id == self.inst_id:
                    return eip
        return None

    def new_eip(self):
        ''' Allocates a new EIP address to an EC2 instance '''
        logging.debug('Allocating a new EIP!', self.verbose)
        self.allocateeips = False
        try:
            eip = self.conn.allocate_address()
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None
        try:
            return eip.associate(self.inst_id)
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None

    def re_associate_eip(self, steal=False, eip=None):
        ''' Reassociates an EIP address to an EC2 instance '''
        if self.node_db:
            logging.debug('Checking for EIP', self.verbose)
            eip = self.get_eip(steal)
        else:
            logging.info('No external data source is defined, skipping')
        if eip:
            logging.debug(
                'Will re-associate EIP %s to %s' %
                (eip.public_ip, self.inst_id), self.verbose)
            self.reeip = eip
        else:
            self.reeip = None

    def destroy_eip(self):
        ''' Destroy EIP address associated with an EC2 instance '''
        eip = self.get_associated_eip()
        if eip:
            logging.info('Disassociating EIP from %s' % self.name)
            try:
                eip.disassociate()
            except EC2ResponseError, msg:
                logging.error(msg.message)
                return None
            tries = 1
            while tries <= 10:
                try:
                    eip.release()
                    logging.debug(
                        'EIP released on try %d' % tries,
                        self.verbose)
                    return True
                except EC2ResponseError, msg:
                    logging.error(msg.message)
                    return None
                tries += 1
                time.sleep(5)
            return False
        else:
            logging.debug('No EIP to delete', self.verbose)
            return True

    def destroy(self):
        ''' Teardown and terminate an EC2 Instance '''
        logging.debug(
            'Checking to see if we need to destory EIP', self.verbose)
        if self.destroy_eip():
            logging.debug('Terminating %s' % self.name, self.verbose)
            try:
                self._instance.terminate()
            except EC2ResponseError, msg:
                logging.error(msg.message)
                return None
            return True
        else:
            logging.error('Unable to destroy EIP, cancel terminating instance')
            return False

    def hibernate(self):
        ''' Hibernate an EC2 instance '''
        if self.get_state() == 'running':
            name_tag = self._instance.tags.get('Name')
            try:
                self._instance.stop()
            except EC2ResponseError, msg:
                logging.error(msg.message)
                return None
            logging.debug(
                'stopped (%s: %s) state: %s' %
                (self.inst_id, name_tag, self.get_state()), self.verbose)
            if self.node_db:
                return self.set_state('hibernating')
            else:
                return True

    def wake(self):
        ''' Start an EC2 instance that is stopped '''
        if self.get_state() == 'stopped':
            name_tag = self._instance.tags.get('Name')
            logging.debug('Attempt starting instance %s' %
                          name_tag, self.verbose)
            self.re_associate_eip()
            try:
                self._instance.start()
            except EC2ResponseError, msg:
                logging.error(msg.message)
                return None
            logging.debug(
                'Start succeeded (%s: %s) state: %s' %
                (self.inst_id, name_tag, self.get_state()), self.verbose)
            if self.node_db:
                return self.set_state('inservice')
            else:
                return True



# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
'''
Utilities to deal with CGM-standard accounts
'''

import os
import sys
from kloudi.utils import logging


def get_account_name(account_id=None, mapfile=None):
    '''
    Returns the canonical name of the account passed in, or defaults to the
    account in the environment variable 'AWS_ACCOUNT_ID'
    '''
    if account_id is None:
        account_id = os.getenv('AWS_ACCOUNT_ID')
    if not account_id:
        return False
    return translate_account(account_id=account_id, mapfile=mapfile)


def get_account_id(account_name=None, mapfile=None):
    ''' Returns the AWS account ID. '''
    if account_name is None:
        return os.getenv('AWS_ACCOUNT_ID')
    else:
        return translate_account(account_name=account_name, mapfile=mapfile)


def translate_account(account_id=None, account_name=None, mapfile=None):
    ''' Translate account id to account name and back again. '''
    if not mapfile:
        mapfile = '%s/%s' % (
            os.getenv('AWS_CONF_DIR'), 'account_aliases_map.txt')
    try:
        mfile = open(mapfile, 'r')
    except IOError, msg:
        logging.error(msg)
        return False
    for line in mfile.readlines():
        if len(line.split(':')) == 3:
            (aws_name, aws_id) = line.split(":")[-2:3]
            if account_id:
                if aws_id.rstrip() == account_id:
                    return aws_name.rstrip()
            elif account_name:
                if aws_name.rstrip() == account_name:
                    return aws_id.rstrip()
    if account_id:
        msg = "Couldn't translate ID %s" % account_id
    elif account_name:
        msg = "Couldn't translate account_name %s" % account_name
    logging.error(msg)
    return None



''' Manage route53 hosted zones '''
# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import re
from boto.route53 import exception
from boto.route53.record import ResourceRecordSets


def verbose_msg(message, verbose):
    ''' Function for verbose output. '''
    if verbose:
        print 'DEBUG: %s' % message


class ZoneMetrics(object):
    ''' Class to gather route53 metrics '''
    def __init__(self):
        self.updates = 0
        self.creates = 0
        self.count = 0

    def metric_count(self, metric):
        ''' Increment metric count. '''
        if metric == 'updates':
            self.updates += 1
        if metric == 'creates':
            self.creates += 1

    def get_metrics(self):
        ''' Returns route53 metrics '''
        return (self.creates, self.updates)


class HostedZone(object):
    ''' Class to manage a route53 domain '''
    def __init__(self, conn, account, comment=None,
                 domain="ctgrd.com", verbose=False):
        self.conn = conn
        self.domain = "%s.%s" % (account, domain)
        self.comment = comment
        self.verbose = verbose
        self.zone_id = self.get_zone_id(self.domain)
        self.re_uqdn = re.compile("." + self.domain + ".?$")
        self.changes = ResourceRecordSets(
            self.conn, self.zone_id, self.comment)
        self.metrics = ZoneMetrics()

    def get_zone_id(self, domain):
        ''' Returns the zone id for the given domain.
        >>> dns.zone_id
        u'Z35T6NURAMOZT5'
        '''
        zones = self.conn.get_all_hosted_zones()
        domain = domain + '.'
        for zone in zones['ListHostedZonesResponse']['HostedZones']:
            msg = 'Found zone: %s' % zone['Name']
            verbose_msg(msg, self.verbose)
            if zone['Name'] == domain:
                msg = ('Using zone: %s domain: %s zoneId: %s' %
                       (zone['Name'], domain, zone['Id'].split('/')[2]))
                verbose_msg(msg, self.verbose)
                return zone['Id'].split("/")[2]
        print "ERROR Couldn't find domain %s" % domain
        return None

    def get_fqdn(self, name):
        ''' Retruns the FQDN for the host name given.
        >>> dns.get_fqdn('faketest1')
        'faketest1.opsqa.ctgrd.com'
        >>> dns.get_fqdn('faketest2.opsqa.ctgrd.com')
        'faketest2.opsqa.ctgrd.com'
        '''
        if self.domain == name[-len(self.domain):]:
            return name
        else:
            return "%s.%s" % (name, self.domain)

    def add_record(self, name, record):
        ''' Prepares to add new records.
        >>> dns.changes.changes = list()
        >>> nodes = dict()
        >>> nodes['faketest1.opsqa.ctgrd.com'] = {'type': 'CNAME',
        ...                                       'ttl': 60,
        ...                                       'resource': ['dummyhost1']}
        >>> nodes['faketest2.opsqa.ctgrd.com'] = {'type': 'CNAME',
        ...                                       'ttl': 60,
        ...                                       'resource': ['dummyhost2']}
        >>> dns.add_record('faketest1', nodes['faketest1.opsqa.ctgrd.com'])
        >>> dns.add_record('faketest2', nodes['faketest2.opsqa.ctgrd.com'])
        >>> dns.changes.changes
        [['CREATE', <Record:faketest1.opsqa.ctgrd.com:CNAME:dummyhost1>], \
['CREATE', <Record:faketest2.opsqa.ctgrd.com:CNAME:dummyhost2>]]
        '''
        if self.zone_id is None:
            print "Zone ID not set"
            return False
        change = self.changes.add_change(
            "CREATE", self.get_fqdn(name), record['type'], record['ttl'])
        for rec in record['resource']:
            change.add_value(rec)
        verbose_msg('Added record: %s' % change, self.verbose)
        self.metrics.metric_count('creates')

    def delete_record(self, name, record):
        ''' Prepares to delete existing records.
        >>> dns.changes.changes = list()
        >>> nodes = dict()
        >>> nodes['faketest1.opsqa.ctgrd.com'] = {'type': 'CNAME',
        ...                                       'ttl': 60,
        ...                                       'resource': ['dummyhost1']}
        >>> nodes['faketest2.opsqa.ctgrd.com'] = {'type': 'CNAME',
        ...                                       'ttl': 60,
        ...                                       'resource': ['dummyhost2']}
        >>> dns.delete_record('faketest1', nodes['faketest1.opsqa.ctgrd.com'])
        >>> dns.delete_record('faketest2', nodes['faketest2.opsqa.ctgrd.com'])
        >>> dns.changes.changes
        [['DELETE', <Record:faketest1.opsqa.ctgrd.com:CNAME:dummyhost1>], \
['DELETE', <Record:faketest2.opsqa.ctgrd.com:CNAME:dummyhost2>]]
        '''
        if self.zone_id is None:
            print "Zone ID not set"
            return False
        change = self.changes.add_change(
            "DELETE", self.get_fqdn(name), record['type'], record['ttl'])
        for rec in record['resource']:
            change.add_value(rec)

    def update_record(self, name, record):
        ''' Prepares to update existing records.
        >>> dns.changes.changes = list()
        >>> nodes = dict()
        >>> nodes['faketest1.opsqa.ctgrd.com'] = {'type': 'CNAME',
        ...                                       'ttl': 60,
        ...                                       'resource': ['dummyhost1']}
        >>> nodes['faketest2.opsqa.ctgrd.com'] = {'type': 'CNAME',
        ...                                       'ttl': 60,
        ...                                       'resource': ['dummyhost2']}
        >>> dns.update_record('faketest1', nodes['faketest1.opsqa.ctgrd.com'])
        >>> dns.update_record('faketest2', nodes['faketest2.opsqa.ctgrd.com'])
        >>> dns.changes.changes
        [['UPSERT', <Record:faketest1.opsqa.ctgrd.com:CNAME:dummyhost1>], \
['UPSERT', <Record:faketest2.opsqa.ctgrd.com:CNAME:dummyhost2>]]
        '''
        if self.zone_id is None:
            print "Zone ID not set"
            return False
        change = self.changes.add_change(
            "UPSERT", self.get_fqdn(name), record['type'], record['ttl'])
        for rec in record['resource']:
            change.add_value(rec)
        self.metrics.metric_count('updates')

    def commit_records(self):
        ''' Commit route53 records to aws API. '''
        msg = 'Submitted Records:\n\t%s ' % self.changes.changes
        verbose_msg(msg, self.verbose)
        try:
            self.changes.commit()
            return True
        except exception.DNSServerError, error:
            print 'ERROR: %s' % error.message
            return False

    def add_update_records(self, nodes, record_type=None, ttl=60):
        ''' Batch CREATE and UPSERT Route53 records.
        >>> dns.changes.changes = list()
        >>> nodes = dict()
        >>> nodes['faketest1.opsqa.ctgrd.com'] = {'type': 'CNAME',
        ...                                       'ttl': 60,
        ...                                       'resource': ['dummyhost1']}
        >>> nodes['faketest2.opsqa.ctgrd.com'] = {'type': 'CNAME',
        ...                                       'ttl': 60,
        ...                                       'resource': ['dummyhost2']}
        '''
        #>>> dns.add_update_records(nodes, record_type='CNAME', ttl=60)
        #True
        (exrecords, nrecords, del_records) = self.generate_records(
            nodes, record_type=record_type, ttl=ttl)
        if exrecords:
            print 'INFO: Records that match route53: %s' % len(exrecords)
            for name in exrecords:
                ovalue = exrecords[name]['resource']
                value = nodes[name.strip('.')]['resource']
                if ovalue != value:
                    print ("INFO: Updating, %s: %s" %
                           (name, nodes[name.strip('.')]))
                    self.update_record(name.strip('.'), nodes[name.strip('.')])
        if nrecords:
            print 'INFO: New records: %s' % len(nrecords)
            for name in nrecords:
                print "INFO: Adding %s: %s" % (name, nrecords[name]['resource'])
                self.add_record(name.strip('.'), nrecords[name])
        if del_records:
            print 'INFO: Record(s) to delete: %s' % len(del_records)
            for name in del_records:
                print 'INFO: To Delete: %s: %s' % (name, del_records[name])
        if len(self.changes.changes) != 0:
            print 'INFO: Submitting record changes to route53'
            return self.commit_records()
        else:
            print 'INFO: No changes to be made for: %s' % self.domain
            return False

    def get_metrics(self):
        ''' Return route53 metrics
        >>> dns.get_metrics() # doctest: +ELLIPSIS
        INFO: Records in route53:...
        '''
        (creates, updates) = self.metrics.get_metrics()
        records = self.get_records(record_type='CNAME', name=self.domain)
        return (creates, updates, len(records))

    def get_records(self, record_type=None, name=None, maxitems=None):
        ''' Get all records for given params.
        >>> records = dns.get_records() # doctest: +ELLIPSIS
        INFO: Records in route53:...
        >>> records[0] # doctest: +ELLIPSIS
        <Record:opsqa.ctgrd.com.:NS:ns-430.awsdns-53.com...
        >>> records = dns.get_records(
        ...     record_type='CNAME',
        ...     name='opsqa.ctgrd.com.') # doctest: +ELLIPSIS
        INFO: Records in route53:...
        >>> records['aws1devjmp1.opsqa.ctgrd.com.']
        {'resource': [u'ec2-50-16-187-199.compute-1.amazonaws.com'], \
'type': u'CNAME', \
'ttl': u'60'}
        '''
        if self.zone_id is None:
            print "Zone ID not set"
            return False
        srec = dict()
        records = self.conn.get_all_rrsets(self.zone_id, type=record_type,
                                           name=name, maxitems=maxitems)
        if record_type is not None:
            for rec in records:
                if rec.type == record_type:
                    srec[rec.name] = {'type': rec.type,
                                      'ttl': rec.ttl,
                                      'resource': rec.resource_records}
            print 'INFO: Records in route53: %s' % len(srec)
            return srec
        else:
            print 'INFO: Records in route53: %s' % len(records)
            return records

    def generate_records(self, nodes, record_type=None, ttl=60):
        ''' Check to see if record exists. Returns new and existing records.
        >>> nodes = {
        ...     'faketest1.opsqa.ctgrd.com': {
        ...         'resource': ['dummyhost1'],
        ...         'ttl': 60,
        ...         'type': 'CNAME'},
        ...     'faketest2.opsqa.ctgrd.com': {
        ...         'resource': ['dummyhost2'],
        ...         'ttl': 60,
        ...         'type': 'CNAME'}}
        >>> (erecords, nrecords, drecords) = dns.generate_records(
        ...  nodes, record_type='CNAME') # doctest: +ELLIPSIS
        INFO: Records in route53:...
        INFO: Record generation complete
        >>> erecords
        {}
        >>> nrecords
        {'faketest2.opsqa.ctgrd.com.': {'resource': ['dummyhost2'], \
'type': 'CNAME', \
'ttl': 60}, \
'faketest1.opsqa.ctgrd.com.': {'resource': ['dummyhost1'], \
'type': 'CNAME', \
'ttl': 60}}
        >>> type(drecords)
        <type 'dict'>
        '''
        exrecords = dict()
        new_records = dict()
        del_records = dict()
        records = self.get_records(name=self.domain, record_type=record_type)
        for name in nodes:
            name = self.get_fqdn(name) + "."
            if name in records:
                if record_type is not None:
                    if records[name]['type'] == record_type:
                        exrecords[name] = records[name]
                    else:
                        print ('WARN: Record %s found,'
                               'but type is not the same' % name)
                else:
                    exrecords[name] = records[name]
            else:
                new_records[name] = {
                    'type': record_type,
                    'ttl': ttl,
                    'resource': nodes[name.strip('.')]['resource']}
        ignore_records = 0
        unmanaged = ['rds.amazonaws.com', 'elb.amazonaws.com', 'amazonses.com']
        for name in records:
            if name not in exrecords:
                if (any(condition in records[name]['resource']
                        for condition in unmanaged)):
                    print 'INFO: Ignoring unmanaged record %s' % name
                    ignore_records += 1
                else:
                    del_records[name] = records[name]
        if ignore_records != 0:
            print 'INFO: Ignoring %s records, unmanaged' % ignore_records
        print 'INFO: Record generation complete'
        return (exrecords, new_records, del_records)

    def print_records(self, record_type=None, name=None,
                      maxitems=None, nms=True, soa=False):
        ''' Print Records found in Route53.
        >>> dns.print_records(record_type='NS', name='opsqa.ctgrd.com')
        INFO: Records in route53: 1
        Name             Type  TTL     Value
        opsqa.ctgrd.com. NS    172800  ns-430.awsdns-53.com.
        opsqa.ctgrd.com. NS    172800  ns-1157.awsdns-16.org.
        opsqa.ctgrd.com. NS    172800  ns-620.awsdns-13.net.
        opsqa.ctgrd.com. NS    172800  ns-1859.awsdns-40.co.uk.
        '''
        if self.zone_id is None:
            print "Zone ID not set"
            return False
        records = self.get_records(
            record_type=record_type, name=name, maxitems=maxitems)
        fmt = "%-16s %-5s %-7s %s"
        print fmt % ("Name", "Type", "TTL", "Value")
        if type(records) == dict:
            for name in records:
                if records[name]['type'] == "NS" and nms is False:
                    continue
                if records[name]['type'] == "SOA" and soa is False:
                    continue
                for rcr in records[name]['resource']:
                    print fmt % (self.get_uqdn(name), records[name]['type'],
                                 records[name]['ttl'], rcr)
        else:
            for rec in records:
                if rec.type == "NS" and nms is False:
                    continue
                if rec.type == "SOA" and soa is False:
                    continue
                for rcr in rec.resource_records:
                    print fmt % (self.get_uqdn(rec.name), rec.type,
                                 rec.ttl, rcr)

    def get_uqdn(self, name):
        ''' Returns the UQDN
        >>> dns.get_uqdn('aws1devjmp1.opsqa.ctgrd.com.')
        'aws1devjmp1'
        '''
        return self.re_uqdn.sub("", name)

    def get_tiny_dns_zone(self):
        ''' Export tiny dns zone.
        >>> zone = dns.get_tiny_dns_zone() # doctest: +ELLIPSIS
        INFO:...
        >>> zone # doctest: +ELLIPSIS
        u'&opsqa.ctgrd.com::ns-430.awsdns-53.com...
        '''
        zone = ""
        if self.zone_id is None:
            print "Zone ID not set"
            return False
        records = self.get_records(name=self.domain)
        for rec in records:
            for rcr in rec.resource_records:
                if rec.type == "NS":
                    zone += "&%s::%s\n" % (rec.name.rstrip("."), rcr)
                elif rec.type == "A":
                    zone += "+%s:%s\n" % (rec.name.rstrip("."), rcr)
                elif rec.type == "CNAME":
                    zone += "C%s:%s:%d\n" % (rec.name.rstrip("."),
                                             rcr, int(rec.ttl))
                elif rec.type == "SOA":
                    src = map(lambda x: x.rstrip("."), rcr.split(" "))
                    zone += "Z%s::%s:%s\n" % (rec.name.rstrip("."),
                                              src[0], src[1])
                else:
                    zone += "#%s %s:%s\n" % (rec.type,
                                             rec.name.rstrip("."), rcr)
        return zone



# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
''' This module sends metrics directly to graphite. '''

import re
import socket
import time
from kloudi.utils import logging


class Graphite(object):
    ''' This class sends metrics to graphite. '''
    def __init__(self, server=None, port=2003, debug=False):
        self.server = server
        self.port = port
        self._sock = None
        self._sock_status = None
        self.prefix = None
        self.debug = debug

    def send_metric(self, name, value):
        ''' Send custom path metric to graphite. '''
        message = "%s %s %s"  % (name, value, int(time.time()))
        if self.debug:
            logging.info(message)
            return True
        if not self._sock:
            self.connect()
        if self._sock_status:
            logging.info("Sending %s %s to %s" % (name, value, self.server))
            try:
                self._sock.sendall(message + "\n")
            except socket.gaierror, msg:
                logging.error(msg)
                self._sock = None
                return False

    def send_server_metric(self, name, value, hostname=None):
        ''' Send server path metric to graphite. '''
        message = '%s.%s%s' % (self.get_server_prefix(hostname), name, value)
        if self.debug:
            logging.info(message)
            return True
        return self.send_metric(message)

    def connect(self):
        ''' Create socket connection to graphite. '''
        try:
            self._sock = socket.create_connection(
                (self.server, self.port), timeout=10)
            self._sock_status = True
        except (socket.gaierror, socket.timeout, socket.error), msg:
            self._sock_status = False
            logging.error(msg)

    def get_server_prefix(self, hostname=None):
        ''' Get FQDN for server metric path. '''
        if hostname is None:
            if self.prefix:
                return self.prefix
            hostname = socket.getfqdn()
        fqdn = re.sub(r'\.', '_', hostname)
        ct_class = fqdn[7:10]
        self.prefix = "servers.%s.%s" % (ct_class, fqdn)
        return self.prefix



# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
''' Organize options from config files '''
import os
import re
import sys
import ConfigParser
import kloudi.aws.ec2.ebs
import kloudi.utils.naming
import kloudi.utils.accounts
from kloudi.utils import logging
from ConfigParser import SafeConfigParser


def get_account_from_conf(conf=None):
    ''' Loads config only so we can get the account for ClusterConfig. '''
    scp = SafeConfigParser()
    scp.read(conf)
    try:
        return scp.get('cluster', 'aws_account')
    except (ConfigParser.NoSectionError, ConfigParser.NoOptionError), msg:
        logging.error(msg)
        return None


def get_hvm_instances():
    ''' HVM instance types that are not compatible with paravirtual AMIs '''
    instances = [
        'cc2.8xlarge',
        'i2.xlarge',
        'i2.2xlarge',
        'i2.4xlarge',
        'i2.8xlarge',
        'r3.large',
        'r3.xlarge',
        'r3.2xlarge',
        'r3.4xlarge',
        'r3.8xlarge',
        't2.micro',
        't2.small',
        't2.medium',
    ]
    return instances


def verify_az(avz):
    ''' Verify AZ via regex '''
    if re.match(r"^\w+-\w+-\d\w$", avz):
        return True
    return False


class TooManyAMIsError(Exception):
    ''' Returns too many AMI exception '''
    def __init__(self, value):
        super(TooManyAMIsError, self).__init__(value)
        self.value = value

    def __str__(self):
        return self.value


class AMINotFoundError(Exception):
    ''' Return AMI not found exception '''
    def __init__(self, value):
        super(AMINotFoundError, self).__init__(value)
        self.value = value

    def __str__(self):
        return self.value


class InvalidAZError(Exception):
    ''' Returns invalid AZ exception'''
    def __init__(self, value):
        super(InvalidAZError, self).__init__(value)
        self.value = value

    def __str__(self):
        return self.value


class InvalidCIDRNameError(Exception):
    ''' Return invalid cidr name exception '''
    def __init__(self, value):
        super(InvalidCIDRNameError, self).__init__(value)
        self.value = value

    def __str__(self):
        return self.value


class ConfigNotFoundException(Exception):
    ''' Return config not found exception '''
    def __init__(self, value):
        super(ConfigNotFoundException, self).__init__(value)
        self.value = value

    def __str__(self):
        return self.value

class EBSConfig(object):
    ''' A class to hold EBS configuration data '''
    def __init__(self):
        self.volumes = list()
        self.azs = list()

    def add_volumes(self, vol_type, device, size, iops):
        ''' Add volume information '''
        self.volumes.append(
            {'type': vol_type, 'device': device, 'size': size, 'iops': iops})

    def get_volumes(self):
        ''' Return volume information '''
        return self.volumes

    def set_azs(self, azs):
        ''' Set AZ for volume '''
        self.azs = azs

    def get_azs(self):
        ''' Get AZ information for volume '''
        return self.azs


class ELBConfig(object):
    ''' A simple class to hold ELB configuration data '''
    # pylint: disable=too-many-instance-attributes
    # Appropriate number of attributes for an ELB
    def __init__(self):
        self.enabled = None
        self.protocol = None
        self.public_port = None
        self.private_port = None
        self.vip_number = None
        self.hc_access_point = None
        self.hc_interval = None
        self.hc_target = None
        self.hc_healthy_threshold = None
        self.hc_unhealthy_threshold = None
        self.azs = list()

    def validate(self):
        ''' Validate required config options are set'''
        if not self.enabled:
            return self.enabled
        items = [
            'protocol', 'public_port', 'private_port',
            'hc_access_point', 'hc_interval', 'hc_target',
            'hc_healthy_threshold', 'hc_unhealthy_threshold', 'vip_number']
        for item in items:
            if getattr(self, item) is None:
                msg = '%s not set, disabling ELB' % item
                logging.error(msg)
                self.enabled = False
        return self.enabled

    def set_azs(self, azs):
        ''' Set ELB AZ '''
        self.azs = azs

    def get_azs(self):
        ''' Set ELB AZ '''
        return self.azs


class SGConfig(object):
    ''' A class to store SG config objects '''
    def __init__(self):
        self.cidr_rules = list()
        self.sg_rules = list()

    def add_cidr(self, proto, fport, lport, cidr):
        ''' Add cidr rules '''
        self.cidr_rules.append(
            {'proto': proto, 'fport': fport, 'lport': lport, 'cidr': cidr})

    def add_sg(self, proto, fport, lport, owner, sgrp):
        ''' Add SG rules '''
        # pylint: disable=too-many-arguments
        # Appropriate number arguments of for add_sg
        self.sg_rules.append(
            {'proto': proto, 'fport': fport,
             'lport': lport, 'owner': owner, 'sg': sgrp})

    def get_cidr(self):
        ''' Get cidr rules '''
        return self.cidr_rules

    def get_sg(self):
        ''' Get sg rules '''
        return self.sg_rules


class RAIDConfig(object):
    ''' A simple class to hold RAID configuration data '''
    def __init__(self):
        self.enabled = None
        self.level = None
        self.device = None

    def set_level(self, level):
        ''' Set the RAID level '''
        self.level = level

    def set_device(self, device):
        ''' Set the RAID device '''
        self.device = device

    def get_level(self):
        ''' Returns RAID level '''
        return self.level

    def get_device(self):
        ''' Returns RAID device '''
        return self.device


class RDSSGConfig(object):
    ''' This Class is used to build RDS SG Authorizations. '''
    def __init__(self):
        self.cidr_rules = list()
        self.sg_rules = list()

    def add_cidr(self, cidr):
        ''' Add CIDR rules to cidr list. '''
        self.cidr_rules.append({'cidr': cidr})

    def add_sg(self, oid, sid):
        ''' Add SG rules to cidr list. '''
        self.sg_rules.append({'oid': oid, 'sid': sid})

    def get_cidr(self):
        ''' Returns populated list of CIDR rules. '''
        return self.cidr_rules

    def get_sg(self):
        ''' Returns populated list of SG rules. '''
        return self.sg_rules


class RDSPGConfig(object):
    ''' This Class is used to get RDS Parameters. '''
    def __init__(self):
        self.rds_parameters = list()

    def add_parameter(self, parameter):
        ''' Add RDS Parameter to list. '''
        self.rds_parameters.append(parameter)

    def get_parameters(self):
        ''' Returns list of paramters. '''
        return self.rds_parameters


class RDSInstanceConfig(object):
    ''' Manages storeing RDS config items. '''
    def __init__(self):
        self.rds_conf = dict()

    def add_config(self, key, value):
        ''' Used to Store RD config items into a dictionary. '''
        self.rds_conf[key] = value

    def get_config(self):
        ''' Returns RDS config dictionary. '''
        return self.rds_conf


class ClusterConfig(object):
    """ A class to hold and manage cluster configuration data
    Config will come from all of the following, in priority order:
    1) The command line, via the Set methods
    2) The Class config file, named $env$class.ini
    3) The $AWS_CONF_DIR/cluster_defaults.ini.$AWS_PROFILE_NAME
    4) The $AWS_CONF_DIR/cluster_defaults.ini
    5) The $HOME/.cluster_defaults.ini (for "sshkey" only)
    """
    # pylint: disable=too-many-instance-attributes
    # Appropriate number of attributes for ClusterConfig
    def __init__(self, ini_file=None, account_name=None, prv_type='ec2',
                 verbose=False, no_defaults=False):
        self.no_defaults = no_defaults  # only read self.classfile if True
        self.defaults = os.getenv('AWS_CONF_DIR') + '/cluster_defaults.ini'
        # This should hold only the ssh key
        self.personal_defaults = os.getenv('HOME') + '/.cluster_defaults.ini'
        self.classfile = ini_file
        if not os.path.exists(ini_file):
            raise ConfigNotFoundException('Invalid config: %s' % ini_file)
        self.verbose = verbose
        self.account_name = account_name
        self.global_ssg = None
        self.primary_sg = None
        # These aren't IN the config file, they're implied by the name
        self.server_env = None
        self.server_class = None
        # This should be looked up from the AZs used
        self.server_datacenter = None
        self.user_data_raw = None
        self.domain = None
        self.ebs = EBSConfig()
        self.elb = ELBConfig()
        self.sgrp = SGConfig()
        self.raid = RAIDConfig()
        self.rds = RDSInstanceConfig()
        self.rds_sg = RDSSGConfig()
        self.rds_pg = RDSPGConfig()
        self.tagset = dict()
        self.overrides = dict()
        self.ini_files = list()
        self.ini = None
        # Read the in the INI files
        if self.no_defaults:
            self.read_files([self.classfile])
        else:
            self.read_files(
                [self.personal_defaults, self.defaults,
                 "%s-%s" % (self.defaults, self.account_name), self.classfile])
        self.read_sections(prv_type)

    def read_sections(self, prv_type):
        ''' Read sections based on provisioning type '''
        if prv_type == 'ec2':
            self.get_meta_data()
            self.server_datacenter = self.get_cg_region()
            if self.ini.has_section('ebs'):
                self.read_ebs_config()
            if self.ini.has_section('elb'):
                self.read_elb_config()
            if self.ini.has_section('securitygroup'):
                self.read_sg_config()
            if self.ini.has_section('raid'):
                self.read_raid_config()
        elif prv_type == 'rds':
            self.get_meta_data()
            self.server_datacenter = self.get_cg_region()
            if self.ini.has_section('rds_provision'):
                self.read_rds_config()
            if self.ini.has_section('rds_securitygroup'):
                self.read_rds_sg_config()
            if self.ini.has_section('rds_parameters'):
                self.read_rds_pg_config()
        elif prv_type == 's3':
            pass

    def get_meta_data(self):
        ''' Get metadata from classfile '''
        self.server_env = os.path.basename(self.classfile)[:3]
        self.server_class = os.path.basename(self.classfile)[3:6]
        self.primary_sg = '%s%s' % (self.server_env, self.server_class)
        self.global_ssg = 'ssg-management'

    def get_server_env(self):
        ''' Return the server Environment '''
        return self.server_env

    def get_primary_sg(self):
        ''' Return the primary SG '''
        return self.primary_sg

    def get_global_ssg(self):
        ''' Return the Global SG '''
        return self.global_ssg

    def get_aws_region(self):
        ''' We work in only one region, so we can just take the first '''
        if self.get_azs()[0] == 'auto':
            return 'us-east-1'
        else:
            return self.get_azs()[0][:-1]

    def get_cg_region(self):
        ''' Return region from kloudi.utils.naming.get_aws_dc '''
        return kloudi.utils.naming.get_aws_dc(self.get_aws_region())

    def read_files(self, conf_files):
        ''' Read in ini files '''
        logging.debug('Trying %s' % conf_files, self.verbose)
        for ini in conf_files:
            if os.path.exists(ini):
                self.ini_files.append(ini)
        logging.debug('Reading %s' % self.ini_files, self.verbose)
        self.ini = ConfigParser.ConfigParser({
            'AWS_BASE_DIR': os.getenv('AWS_BASE_DIR'),
            'AWS_CONF_DIR': os.getenv('AWS_CONF_DIR')})
        self.ini.read(self.ini_files)

    def get_ini(self, section, name, castf, fallback=None):
        ''' Get a setting from the ini files '''
        try:
            return castf(self.ini.get(section, name))
        except ConfigParser.NoSectionError, msg:
            logging.error(msg)
            return fallback
        except ConfigParser.NoOptionError, msg:
            logging.error(msg)
            return fallback
        return fallback

    def set_ami(self, ami):
        ''' Set AMI '''
        self.overrides['ami'] = ami

    def get_ami(self):
        ''' Return the AMI '''
        if 'ami' in self.overrides:
            return self.overrides['ami']
        instance_type = self.get_size()
        raw_ami = self.get_ini('cluster', 'ami', str)
        if raw_ami.count('VTYPE'):
            if instance_type in get_hvm_instances():
                return raw_ami.replace('VTYPE', 'hvm')
            else:
                return raw_ami.replace('VTYPE', 'paravirtual')
        else:
            return raw_ami

    def get_whitelist_url(self):
        ''' Return the whitelist URL for puppet whitelisting '''
        if 'whitelisturl' in self.overrides:
            return self.overrides['whitelisturl']
        return self.get_ini("cluster", "whitelisturl", str)

    def get_resolved_ami(self, node_db):
        ''' Return resolved AMI '''
        ami = self.get_ami()
        if ami[:4] == 'ami-':
            logging.error(
                'AMI statically set to %s. Please use graffiti values' % ami)
            return ami
        try:
            amis = node_db.get_amis(self.get_cg_region(), ami)
        except:
            raise AMINotFoundError("No AMI matching '%s' found" % ami)
        if amis is None:
            raise AMINotFoundError("No AMI matching '%s' found" % ami)
        if len(amis) == 1:
            newami = amis.values()[0]
            self.set_ami(newami)
            logging.debug(
                "Converted '%s' to '%s'" % (ami, newami), self.verbose)
            return newami
        elif len(amis) > 1:
            raise TooManyAMIsError("%s matches too many AMIs: %s" % (ami, amis))

    def limit_azs(self, limit):
        ''' Limit the number of AZs to use '''
        if limit > 0:
            oldazs = self.get_azs()
            newazs = oldazs[:limit]
            self.set_azs(','.join(newazs))
            return len(oldazs) - len(newazs)
        else:
            logging.error("Trying to limit AZs to %d" % limit)
        return 0

    def set_azs(self, azs):
        ''' Set comma sperated list of AZs '''
        for avz in azs.split(","):
            if not verify_az(avz):
                raise InvalidAZError("AZ '%s' is invalid" % avz)
        self.overrides['azs'] = azs.split(",")

    def get_azs(self):
        ''' Return AZ information '''
        zones = list()
        if 'azs' in self.overrides:
            return self.overrides['azs']
        ret = self.get_ini("cluster", "zone", str)
        if ret:
            for avz in ret.split(","):
                if not verify_az(avz):
                    raise InvalidAZError("AZ '%s' is invalid" % avz)
                zones.append(avz.strip())
        return zones

    def get_next_az(self):
        ''' We'll need them in a list to do this, stick in overrides '''
        if 'azs' not in self.overrides:
            self.overrides['azs'] = self.get_azs()
        avz = self.overrides['azs'].pop(0)
        self.overrides['azs'].append(avz)
        return avz

    def get_count_azs(self):
        ''' Get the count of unique AZs '''
        return len(set(self.get_azs()))

    def set_count(self, count):
        ''' Set the instance count '''
        self.overrides['count'] = int(count)

    def get_count(self):
        ''' Return the instance count '''
        if 'count' in self.overrides:
            return self.overrides['count']
        return self.get_ini("cluster", "instance_count", int)

    def set_size(self, size):
        ''' Set the instance size '''
        self.overrides['size'] = size

    def get_size(self):
        ''' Return the instance size '''
        if 'size' in self.overrides:
            return self.overrides['size']
        return self.get_ini("cluster", "instance_size", str)

    def set_ssh_key(self, sshkey):
        ''' Set the ssh key '''
        self.overrides['sshkey'] = sshkey

    def get_ssh_key(self):
        ''' Return the ssh key '''
        if 'sshkey' in self.overrides:
            return self.overrides['sshkey']
        return self.get_ini("ssh", "sshkey", str)

    def get_dc(self):
        ''' Get the AWS region '''
        if self.get_ini("DEFAULT", "datacenter", str) is not None:
            logging.error(
                "The 'datacenter' option is no longer read from the INI file")
        return self.get_cg_region()

    def get_user_data_file(self):
        ''' Return the userdata file '''
        return self.get_ini("cluster", "user_data_file", str, None)

    def get_user_data(self, replacements=None):
        ''' Get userdata and set replacements '''
        path = self.get_user_data_file()
        logging.debug('user_data_file: %s' % path, self.verbose)
        if not self.user_data_raw:
            if os.path.exists(path):
                try:
                    udfile = file(path, "r")
                except IOError, msg:
                    logging.error(msg)
                    return None
                self.user_data_raw = udfile.read()
                udfile.close()
        udata = self.user_data_raw
        if replacements:
            for key in replacements.keys():
                logging.debug(
                    'Replacing %s with %s in %s' %
                    (key, replacements[key], path), self.verbose)
                udata = udata.replace(key, replacements[key])
        return udata.strip()

    def get_tagset(self):
        ''' Return the tagset cost tags '''
        self.tagset['BusinessUnit'] = self.get_ini("tags", "business_unit", str)
        self.tagset['Team'] = self.get_ini("tags", "team", str)
        self.tagset['Project'] = self.get_ini("tags", "project", str)
        if any(ent for ent in self.ini_files if ent.endswith('meta.ini')):
            self.tagset['Component'] = self.get_ini("tags", "component", str)
            self.tagset['Env'] = self.get_ini("tags", "env", str)
        else:
            comp = self.get_ini("tags", "component", str)
            if comp[:4] == self.server_class + ' ':
                self.tagset['Component'] = self.get_ini(
                    "tags", "component", str)
            else:
                self.tagset['Component'] = "%s %s" % (
                    self.server_class, self.get_ini("tags", "component", str))
            self.tagset['Env'] = self.get_server_env()
        return self.tagset

    def get_launch_timeout(self):
        ''' Return launch timeout '''
        return self.get_ini("cluster", "launch_timeout", int)

    def get_sleep_step(self):
        ''' R eturn sleep step '''
        return self.get_ini("cluster", "sleep_step", int)

    def add_sg(self, sgp):
        ''' Adding additional SGs '''
        if 'other_sgs' not in self.overrides:
            self.overrides['other_sgs'] = self.get_additional_sgs()
        self.overrides['other_sgs'].append(sgp)

    def get_additional_sgs(self):
        ''' Returns additonal SGs'''
        other_sgs = list()
        if 'other_sgs' in self.overrides:
            return self.overrides['other_sgs']
        ret = self.get_ini("cluster", "additional_sgs", str)
        if ret:
            for sgr in ret.split(','):
                other_sgs.append(sgr.strip())
        return other_sgs

    def get_sgs(self):
        ''' Return all SGs '''
        ret = self.get_additional_sgs()
        ret.append("%s%s" % (self.server_env, self.server_class))
        return ret

    def get_node_groups(self):
        ''' Return Node groups '''
        node_groups = list()
        if 'node_groups' in self.overrides:
            return self.overrides['node_groups']
        ret = self.get_ini("cluster", "node_groups", str)
        if ret:
            for ngrp in ret.split(','):
                node_groups.append(ngrp.strip())
        return node_groups

    def set_allocate_eips(self):
        ''' Set allocated EIPs '''
        self.overrides['allocate_eips'] = True
        return True

    def get_allocate_eips(self):
        ''' Return allocated EIPs '''
        if 'allocate_eips' in self.overrides:
            return self.overrides['allocate_eips']
        if self.get_ini("cluster", "allocate_eip", str) == "True":
            allocate_eips = True
        else:
            allocate_eips = False
        return allocate_eips

    def set_use_ebs_optimized(self):
        ''' Set use EBS optimized '''
        self.overrides['use_ebs_optimized'] = True

    def get_use_ebs_optimized(self):
        ''' Get EBS optimized option '''
        if 'use_ebs_optimized' in self.overrides:
            return self.overrides['use_ebs_optimized']
        if self.get_ini("cluster", "use_ebs_optimized", str):
            use_ebs_optimized = True
        else:
            use_ebs_optimized = False
        return use_ebs_optimized

    def get_aws_account(self):
        ''' Returns AWS account name. '''
        return self.account_name

    def get_domain(self):
        ''' Returns domain '''
        return self.get_ini('cluster', 'domain', str)

    def get_fs_type(self):
        ''' Get the filesystem type '''
        return self.get_ini('cluster', 'fs_type', str)

    def read_ebs_config(self):
        ''' Read EBS config options '''
        for vol in self.ini.items("ebs"):
            if len(vol[1].split()) == 3:
                device = vol[0]
                (vol_type, size, iops) = vol[1].split(" ")
                self.ebs.add_volumes(
                    vol_type, "/dev/" + device, size, iops)
            elif len(vol[1].split()) == 2:
                device = vol[0]
                (vol_type, size) = vol[1].split(" ")
                self.ebs.add_volumes(
                    vol_type, "/dev/" + device, size, None)

    def get_ebs_config(self):
        ''' Return EBS config options '''
        return self.ebs.get_volumes()

    def read_elb_config(self):
        ''' Read in ELB config options '''
        if self.get_ini("elb", "enabled", str) == "True":
            self.elb.enabled = True
        else:
            self.elb.enabled = False
            return False
        self.elb.protocol = self.get_ini("elb", "protocol", str)
        self.elb.public_port = self.get_ini("elb", "public_port", int)
        self.elb.private_port = self.get_ini("elb", "private_port", int)
        self.elb.vip_number = self.get_ini("elb", "vip_number", int) or 1
        self.elb.hc_access_point = self.get_ini(
            "healthcheck", "hc_access_point", str)
        self.elb.hc_interval = self.get_ini("healthcheck", "hc_interval", int)
        self.elb.hc_target = self.get_ini("healthcheck", "hc_target", str)
        self.elb.hc_healthy_threshold = self.get_ini(
            "healthcheck", "hc_healthy_threshold", int)
        self.elb.hc_unhealthy_threshold = self.get_ini(
            "healthcheck", "hc_unhealthy_threshold", int)
        self.elb.validate()

    def get_elb_config(self):
        ''' Return ELB config '''
        return self.elb

    def get_elb_name(self):
        ''' Return the name of the ELB, based on cluster and ELB configs '''
        return "%s%svip%s%d" % (
            self.get_cg_region(), self.server_env[:1],
            self.server_class, self.elb.vip_number)

    def read_sg_config(self):
        ''' Reads in SG config options '''
        for item in self.ini.items("securitygroup"):
            if item[1][:7] == "ingress":
                (rtype, proto, ports, remote) = item[1].split(" ")
                if ports == "None":
                    (prt1, prt2) = [-1, -1]
                elif '-' in ports:
                    (prt1, prt2) = ports.split("-")
                else:
                    prt1 = prt2 = ports
                prt1 = int(prt1)
                prt2 = int(prt2)
                if remote[:5] == 'CIDR:':
                    self.sgrp.add_cidr(proto, prt1, prt2, remote[5:])
                elif remote[:4] == 'Net:':
                    cidr = kloudi.utils.naming.get_cidr(remote[4:])
                    if not cidr:
                        raise InvalidCIDRNameError(
                            "Network '%s' is invalid" % remote[4:])
                    self.sgrp.add_cidr(proto, prt1, prt2, cidr)
                elif remote[:3] == 'SG:':
                    acct, sgrp = remote[3:].split("/")
                    if acct == 'self':
                        acctid = kloudi.utils.accounts.get_account_id(
                            account_name=self.get_aws_account())
                    elif acct == 'amazon-elb':
                        logging.debug('acctid set to amazon-elb', self.verbose)
                        acctid = 'amazon-elb'
                    else:
                        acctid = kloudi.utils.accounts.get_account_id(
                            account_name=acct)
                        logging.debug('%s == %s' % (acct, acctid), self.verbose)
                    if acctid:
                        self.sgrp.add_sg(proto, prt1, prt2, acctid, sgrp)
                    else:
                        logging.error("Can't find my own account.")
                logging.debug(
                    "Allowing %s %s for ports %d to %d from %s" %
                    (rtype, proto, prt1, prt2, remote), self.verbose)

    def get_sg_rules(self):
        ''' Return SG rules '''
        return self.sgrp.get_sg()

    def get_cidr_rules(self):
        ''' Return CIDR rules '''
        return self.sgrp.get_cidr()

    def read_raid_config(self):
        ''' Read in RAID config options '''
        if self.get_ini("raid", "enabled", str) == 'True':
            self.raid.enabled = True
        else:
            self.raid.enabled = False
            return False
        self.raid.set_level(self.get_ini("raid", "level", str))
        self.raid.set_device(self.get_ini("raid", "device", str))

    def read_rds_sg_config(self):
        ''' Reads RDS SG authorizations from ini files. '''
        for rule in self.ini.items('rds_securitygroup'):
            if re.match('.*rule', rule[0]):
                (rtype, rvalue) = rule[1].split(':')
                if rtype == 'Net':
                    cidr = kloudi.utils.naming.get_cidr(rvalue)
                    if cidr:
                        logging.debug('Appending RDS CIDR rule %s' % cidr,
                                    self.verbose)
                        self.rds_sg.add_cidr(cidr)
                elif rtype == 'CIDR':
                    logging.debug('Appending RDS CIDR rule %s' % rvalue,
                                self.verbose)
                    self.rds_sg.add_cidr(rvalue)
                elif rtype == 'SG':
                    (oid, sid) = rvalue.split('/')
                    if oid != 'self':
                        acctid = kloudi.utils.accounts.get_account_id(oid)
                    else:
                        acctid = kloudi.utils.accounts.get_account_id(
                            self.get_aws_account())
                    if acctid:
                        logging.debug(
                            'Appending RDS SG rule %s:%s' % (acctid, sid),
                            self.verbose)
                        self.rds_sg.add_sg(acctid, sid)
                    else:
                        logging.warn("Can't find account for %s" % oid)

    def get_rds_sg_rules(self):
        ''' Returns list of RDS SG rules. '''
        return self.rds_sg.get_sg()

    def get_rds_cidr_rules(self):
        ''' Returns list of RDS CIDR rules. '''
        return self.rds_sg.get_cidr()

    def read_rds_pg_config(self):
        ''' Reads RDS parameters from ini files. '''
        for param in self.ini.items('rds_parameters'):
            if re.match('.*parameter', param[0]):
                (name, value, method) = param[1].split()
                self.rds_pg.add_parameter((name, value, method))

    def get_rds_parameters(self):
        ''' Returns list of RDS parameters. '''
        return self.rds_pg.get_parameters()

    def read_rds_config(self):
        ''' Reads RDS config items from config and store into dictionary. '''
        for item in self.ini.items('rds_provision'):
            if item[1] == 'False':
                self.rds.add_config(item[0], None)
            else:
                self.rds.add_config(item[0], item[1])

    def get_rds_config(self):
        ''' Returns dictonary of RDS config items. '''
        return self.rds.get_config()



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()
CHANGELOG = open(os.path.join(here, 'CHANGELOG.rst')).read()

requires = [
    'boto==2.38.0',
    'moto==0.4.24',
    'Zambi==1.0',
    'nvlib==1.0',
    'ipaddr==2.1.11',
    ]

setup(
    name='kloudi',
    package=['kloudi'],
    version='0.2',
    description='kloudi - Provisions AWS services from easy to read ini configs',
    long_description=README + '\n\n' + CHANGELOG,
    classifiers=[
      "Programming Language :: Python",
      "Programming Language :: Python :: 2.6",
      "Programming Language :: Python :: 2.7",
      "Intended Audience :: System Administrators ",
      ],
    scripts=[
      'scripts/kloudiEC2.py',
      'scripts/kloudiS3.py',
      'scripts/kloudinv2route53.py',
    ],
    author='Jesse Escobedo',
    author_email='jesse@moda.bz',
    url='https://github.com/jesco39/kloudi',
    license='Apache',
    keywords='kloudi C3 AWS automation cloud devops',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    tests_require=requires,
    test_suite='nose.collector',
)



# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
""" This module is used for interfacig with AWS EBS """
from time import sleep
from kloudi.utils import logging
from boto.exception import EC2ResponseError


class KloudiEBS(object):
    """ This class is used to manage EBS """
    def __init__(self, conn):
        self.conn = conn

    def create_volume(self, ebs_size, zone, ebs_type='standard', iops=None):
        """
        Create a volume within a given zone. Support for EBS IOPS is included
        """
        vol = list()
        try:
            vol = self.conn.create_volume(
                ebs_size,
                zone,
                snapshot=None,
                volume_type=ebs_type,
                iops=iops)
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None
        return vol

    def delete_volume(self, volid):
        """ Delete unnattached EBS volume """
        try:
            self.conn.delete_volume(volid)
            return True
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None

    def attach_volume(self, vol_id, instance_id, ebs_device):
        """ Attach a volume to an instance that was created """
        try:
            self.conn.attach_volume(vol_id, instance_id, ebs_device)
            return True
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None

    def detach_volume(self, vol_id, instance_id, ebs_device):
        """ Detach a volume to from an instance """
        try:
            self.conn.detach_volume(vol_id, instance_id, ebs_device)
            return True
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None

    def set_ebs_del_on_term(self, instance_id, device):
        """ Set volumes to delete upon termination """
        mvolume = [device + '=true']
        logging.info(
            "Setting delete attribute for %s on %s" %
            (mvolume, instance_id))
        try:
            self.conn.modify_instance_attribute(
                instance_id, 'blockDeviceMapping', mvolume)
            return True
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None

    def create_snapshot(self, vol_id, desc):
        """ Creates an EBS Snapshot """
        logging.info('Creating Snapshot from %s' % vol_id)
        try:
            snap = self.conn.create_snapshot(vol_id, description=desc)
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None
        logging.info('Snapshot created %s: %s' % (snap.id, desc))
        return snap

    def delete_snapshot(self, snap_id):
        ''' Delete a snapshot '''
        logging.info('Deleting snapshot: %s' % snap_id)
        try:
            return self.conn.delete_snapshot(snap_id)
        except EC2ResponseError, msg:
            logging.error(msg.message)
            return None



#!/usr/bin/python2.6
# Copyright 2015 CityGrid Media, LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
Script to update Route53 with AWS hostnames found in nventory. Initially
works only on AWS hostnames (CNAMEs).
"""

import argparse
from zambi import ZambiConn
from nvlib import Nventory
from kloudi.utils.graphite import Graphite
from kloudi.aws.route53.hostedzone import HostedZone


class Nv2Route53(object):
    ''' Update Route53 with AWS hostnames in nventory. '''
    def __init__(self, options):
        self.options = options
        self.cmgr = ZambiConn()
        self.accounts = self.cmgr.get_accounts(self.options.aws_account)
        for self.account in self.accounts:
            self.dns = None
            self.update_records()
            if self.options.graphite:
                self.send_metrics()

    def send_metrics(self):
        ''' Send route53 metrics to graphite '''
        mpfx = 'business.aws.route53.%s' % self.account
        (creates, updates, count) = self.dns.get_metrics()
        grp = Graphite(server=self.options.graphite_server)
        grp.send_metric(mpfx + ".updates",
                       updates, self.options.debug)
        grp.send_metric(mpfx + ".creates",
                       creates, self.options.debug)
        grp.send_metric(mpfx + ".count",
                       count, self.options.debug)

    def update_records(self):
        ''' Run route53 updates based on nventory '''
        nvd = Nventory(ini_file=self.options.nv_ini)
        conn = self.cmgr.get_connection(self.account, service='route53')
        self.dns = HostedZone(conn, self.account, self.options.comment,
                              domain=self.options.domain)
        data = {
            'loc': '',
            'env': '',
            'sclass': '',
            'acct': self.account,
            'domain': self.options.domain}
        nodes = nvd.get_nodes(data)
        nvec2nodes = dict()
        for node in nodes:
            if node['name'] and node['ec2_instance_id']:
                nvec2nodes[node['name']] = {
                    'type': 'CNAME',
                    'ttl': self.options.ttl,
                    'resource': [node['ec2_public_hostname']]}
        if nvec2nodes:
            self.dns.add_update_records(
                nvec2nodes, record_type='CNAME', ttl=self.options.ttl)
        else:
            print 'INFO: No nodes found in external data source.'


def parser_setup():
    ''' Setup the options parser '''
    desc = 'Manages Route53 from external node classifier. '''
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('-a',
                        action='store',
                        dest='aws_account',
                        help='Use AWS Account')
    parser.add_argument('-d',
                        action='store',
                        dest='domain',
                        help='Use domain, top level tld')
    parser.add_argument('-t',
                        action='store',
                        type=int,
                        dest='ttl',
                        default=60,
                        help='Set Route53 record ttl')
    parser.add_argument('-c',
                        action='store',
                        dest='comment',
                        default='Managed by kloudi-nv2route53.py',
                        help='Set Route53 record comment')
    parser.add_argument('-g',
                        action='store_true',
                        dest='graphite',
                        default=False,
                        help='Send output to graphite')
    parser.add_argument('-G',
                        action='store',
                        dest='graphite_server',
                        default='dev.relay-aws.graphite.ctgrd.com',
                        help='the graphite server to send to')
    parser.add_argument('-D',
                        action='store_true',
                        dest='debug',
                        default=False,
                        help="Print, but don't actually send anything")
    parser.add_argument(
        '-n',
        action='store',
        dest='nv_ini',
        default="/app/secrets/nv_prd.ini",
        help='External DB ini file to use; useful for testing')
    return parser


def main():
    ''' Setup options and call main program '''
    parser = parser_setup()
    options = parser.parse_args()
    Nv2Route53(options)


if __name__ == '__main__':
    main()
